/**********************************************************************
 * Agent.cc
 * Copyright (c) 2001 Virtual Werder
 *
 * implements:
 *   class Agent
 *
 * requires:
 *
 *
 * Created on: 25.03.2001 by wenns
 * Maintainer: harlekin
 * Group:
 * Co-authors: cyco
 *
 * Changelog:
 *
 * $Id: Agent.cc,v 1.23 2004/05/17 11:24:50 harlekin Exp $
 *
 **********************************************************************/

#include "common.hh"
#include "Agent.hh"
#include "Geometry.hh"
#include "debug.hh"
#include "SkillPool.hh"
#include "Logger.hh"
#ifdef DEBUG_ON
#include "AgentPlanLogger.hh"
#include "WorldModelLogger.hh"
#endif
#include "GoalieBehaviour.hh"
#include "WingDefenderBehaviour.hh"
#include "CenterBackBehaviour.hh"
#include "WingMidfielderBehaviour.hh"
#include "CenterMidfielderBehaviour.hh"
#include "WingAttackerBehaviour.hh"
#include "CenterAttackerBehaviour.hh"
#ifdef USE_PLANNER
#include "simpleexecuterfactory.h"
#include "SkillPoolImplVWFactory.h"
#include "./panorama/VwDfgWorldModell.hh"
#endif

extern void  doTerminate();

//global variables
bool   serverAlive = true;

//static members
Effectors    *Agent::effectors    = NULL;
Sensors      *Agent::sensors      = NULL;
Communicator *Agent::communicator = NULL;
WorldModel   *Agent::worldModel   = NULL;
SkillPool    *Agent::skillPool    = NULL;

#ifdef USE_PLANNER
AbsolutePerceptionImplVW *Agent::plannerPerceptionInterface = NULL;
//ActorInterfaceImplVW *Agent::plannerActorInterface = NULL;
Core* Agent::plannerCore = NULL;
bool  Agent::useNetExecutor = false;
#endif


Agent::Agent(const char *serverHost, const int serverPort,
             const char *teamname, const bool goalie){
  communicator = new Communicator(serverHost, serverPort);
  sensors      = new Sensors(communicator, teamname);

  worldModel = new WorldModel();
  worldModel->myself.setGoalKeeper(goalie);
  worldModel->myTeam.setName(teamname);

  ServerCommandHistory  &sc  = worldModel->getSentCommands();
  //ServerCommandHistory  &scb = worldModel->getSentCommands_BackUp();
  ServerCommandCounters &qcc = worldModel->getQueuedCommandCounters();
#ifdef DEBUG_ON
  effectors = new Effectors(communicator, sc, qcc, &logger);
#else
  effectors = new Effectors(communicator, sc, qcc, NULL);
#endif
  effectors->myself = &(worldModel->myself);

  
  if(goalie) effectors->init(teamname, true);   // initialize a goalie
  else       effectors->init(teamname);         // initialize a field player

  skillPool  = new SkillPool(); 
}

Agent::~Agent(){
  delete(behaviour);
#ifdef DEBUG_ON
  delete(logger);
#endif
  delete(worldModel);
  delete(communicator);
  delete(sensors);
  delete(effectors);
  delete(skillPool);
#ifdef USE_PLANNER
  delete(plannerPerceptionInterface);
  //delete(plannerActorInterface);
  delete(plannerCore);
#endif  
}


void Agent::setBehaviour(){
  DEBUG2_HARLEKIN("Agent::setBehaviour(), type = ", behaviourType);
  switch(behaviourType){
    case B_Goalie:
      behaviour = new GoalieBehaviour();
      break;
    case B_WingDefenderLeft:
      behaviour = new WingDefenderBehaviour(true);
      break;
    case B_CenterBackDefenderLeft:
      behaviour = new CenterBackBehaviour(true);
      break;
    case B_CenterBackDefenderRight:
      behaviour = new CenterBackBehaviour(false);
      break;
    case B_WingDefenderRight:
      behaviour = new WingDefenderBehaviour(false);
      break;
    case B_WingMidfielderLeft:
      behaviour = new WingMidfielderBehaviour(true);
      break;
    case B_CenterMidfielder:
      behaviour = new CenterMidfielderBehaviour();
      break;
    case B_WingMidfielderRight:
      behaviour = new WingMidfielderBehaviour(false);
      break;
    case B_WingAttackerLeft:
      behaviour = new WingAttackerBehaviour(true);
      break;
    case B_CenterAttacker:
      behaviour = new CenterAttackerBehaviour();
      break;
    case B_WingAttackerRight:
      behaviour = new WingAttackerBehaviour(false);
      break;
    default:
      behaviour = new Behaviour();
      break;
    }
}
  
#ifdef  USE_PLANNER
void Agent::setPlannerBehaviour(){
  switch(behaviourType){
  case B_Goalie:
    behaviour = new GoalieBehaviour();
    break;
  default:
    behaviour = new PlannerBehaviour();
    break;
  }
}
#endif

/** Initializes the Behaviour for this agent*/
void Agent::init(){
  worldModel->waitForNewInfo();
  worldModel->integrateSensorInfo(sensors);

  int unum = worldModel->myself.getUniformNumber();
  //switch(unum)
  behaviourType = Formations::getCurrentFormation()->getBehaviour(unum);
  //DEBUG2_HARLEKIN("Formations-Testing: Set Behaviour to type: ", type);
#ifdef USE_PLANNER
  setPlannerBehaviour();
#endif

#ifndef USE_PLANNER
  setBehaviour();
#endif
  communicator->sendMessage("(clang (ver 8 8))");
}


void Agent::run(){
  //bool goalie = worldModel->myself.isGoalKeeper();

  //wait for and integrate init in order to have the unum for
  //initializePositioning()
  worldModel->waitForNewInfo();
  worldModel->integrateSensorInfo(sensors);
#ifdef DEBUG_ON
  int unum = worldModel->myself.getUniformNumber();
  // used for worldmodellogging
  int lastCycle = 0;
  int lastCycleStopped = 0;
#endif

#ifdef DEBUG_ON
  logger = new Logger(worldModel);
  AgentPlanLogger::apl  = new AgentPlanLogger( worldModel );
  WorldModelLogger::wml = new WorldModelLogger( worldModel );
#endif

#ifdef USE_PLANNER
  setupPlanner();
#endif

#ifdef DEBUG_ON
  if(unum == ConfigFile::playerToLogWM){
     WorldModelLogger::wml->writeLeadIn();
  }
#endif
      
  while( serverAlive ){

    if( worldModel->integrateSensorInfo(sensors) ){ 
      // returns true if we received a see, sense or fullstate-message
      // or both at the same time. This can happen while playing with
      // synchronzied mode (we will receive all messages from the first 
      // 60 ms at the same time)
      DEBUG2("\t\t\t\tTIME: ", *(worldModel->getTime()));

      if(worldModel->testAndSetNewCycle()) effectors->setNewCycle();
      //else                                 effectors->createBackUp();

      bool isSeeCycle = timerlateras(worldModel->getTimeLastSee(),
				     worldModel->getTimeLastSb() );
      // delete info about sent commands (last cycle)
      if (Configuration::synchMode){
	effectors->setState(ES_Synch);
      }else{
	if (isSeeCycle)
	  effectors->setState(ES_See);
	else
	  effectors->setState(ES_Estimate);
      }
      //effectors->deleteAllCommands();
      //worldModel->getSentCommands().clear();

      //testMajunor();
      //testWenns();
      //testPagra();
      //testTekgeist();
      //testFritsche();
      //if (worldModel->getTime().getTime())
      //testCniehaus();
      //testJosh();

      // normal mode - the thinking phase is activated with each sense or see message
      if ( !Configuration::synchMode ) {
	/*#ifdef USE_PLANNER
	  plannerCore->think();
	  #endif*/
      

	// Hack
	// check if the current formation is equal to the "new" formation which
	int type = Formations::getCurrentFormation()->getBehaviour(worldModel->myself.getUniformNumber());
	if (type != behaviourType){
	  behaviourType = type;
#ifdef USE_PLANNER
	  setPlannerBehaviour();
#else
	  setBehaviour();
#endif
	}
      
	behaviour->generalBehaviour();

	if( isSeeCycle ){
	  //if decisions based on fresh see (will not get better info) ...
	  DEBUG1_SYNCHRO("[REASONING] \trun(): will not get better info, sending" );
	  
	  // ... so send commands
	  effectors->sendAllCommands();
	}
      }
    }

    // else
    // synchmode is active
    if ( Configuration::synchMode ) {
      // the thinking phase is activated by the think message from the server
      if ( worldModel->isNowThinkSegment() ) {

	// Change behaviour if needed (new formation by coach etc.)
	// Hack
	int type = Formations::getCurrentFormation()->getBehaviour(worldModel->myself.getUniformNumber());
	if (type != behaviourType){
	  behaviourType = type;

#ifdef USE_PLANNER
	  setPlannerBehaviour();
#else
	  setBehaviour();
#endif
	}

	// If we arrive here the sense-body message and a potential see message has
	// already been integrated in the worldmodel. Now the agent thinks and sends
	// afterwards his commands and then his "done!". 
	behaviour->generalBehaviour();

	effectors->sendAllCommands();
	communicator->sendMessage( "(done)" );	
      } 
    }

    // wait for new information from the server
    DEBUG1_SYNCHRO("[REASONING] \trun(): waiting for new info" );
    if( worldModel->waitForNewInfo() == false ){
      //comment out the following if debugging with a debugger is wanted
      //otherwise the program exits after n (usually 3) seconds
      serverAlive =  false;
      DEBUG1_SYNCHRO("[REASONING] \trun(): server probably dead" );
    }

#ifdef DEBUG_ON
    // check if this was a new cycle => write the worldmodel to disc
    int thisCycle        = worldModel->getTime()->getTime();
    int thisCycleStopped = worldModel->getTime()->getStoppedCount();
    if(unum == ConfigFile::playerToLogWM && (thisCycle != lastCycle || thisCycleStopped != lastCycleStopped)){
      WorldModelLogger::wml->writeCurrentCycle();
      lastCycle        = thisCycle;
      lastCycleStopped = thisCycleStopped;
    }
#endif
  }

#ifdef DEBUG_ON
  // Before shutting down the agent write the end of the logging file
  if(unum == ConfigFile::playerToLogWM){
    WorldModelLogger::wml->writeLeadOut();
  }
#endif

}

void Agent::testJosh(){ 
}

void Agent::testFritsche(){
}

void Agent::testMajunor(){
}

void Agent::testTekgeist(){
}

void Agent::testPagra(){
}

void Agent::testCniehaus(){
  effectors->queueDash(100);
}

void Agent::testWenns(){
}


#ifdef USE_PLANNER
/** Set up the planner */
void Agent::setupPlanner(){
  // planner initialization
  plannerPerceptionInterface = new AbsolutePerceptionImplVW(worldModel);
  //plannerActorInterface = new ActorInterfaceImplVW();
  plannerCore = new Core();
  plannerCore->setapi(plannerPerceptionInterface);
  //plannerCore->setai(plannerActorInterface);
  plannerCore->setExecutor(new SimpleExecuterFactory());
  if (this->useNetExecutor){
    SkillPoolImplVWFactory spivwf;
    plannerCore->setSkillPool(spivwf.getSkillPool());
  }
  VwDfgWorldModell* vwwm = new VwDfgWorldModell();
  vwwm->setWorldModell(worldModel);
  plannerCore->setWm(vwwm);
}
#endif
