/**********************************************************************
 * ServerCommands.hh
 * Copyright (c) 2002 Virtual Werder
 *
 * implements:
 *
 *   class  ServerCommands
 *   class  ServerCommandHistory
 *   struct ServerCommandCounters
 *
 * requires:
 *
 *
 * Created on: 25.03.2002 by wenns
 * Maintainer: harlekin
 * Group:
 * Co-authors:
 *
 * Changelog:
 *
 * $Id:
 *
 **********************************************************************/

#ifndef SERVERCOMMANDS_HH
#define SERVERCOMMANDS_HH

#include "common.hh"
#include "debug.hh"
#include <iostream>
#include <string.h>

#define SS_MSG_MAX_SIZE  4096 /**< incoming message max size  */
#define MAIN_COMM_SIZE   64   /**< primary commad max size  */
#define OTHER_COMM_SIZE  512  /**< other message max size  */
#define MAX_NO_OTHERS    6    /**< max number of other commands in one cycle */

#define SAY_MSG_MAX_SIZE 512  /**< string size of say message */

/**
 * ServerCommands serves as a container for the soccer server commands. It can
 * store commands (as strings) that can be executed at one timestep.
 * Two kinds of commands are descriminated: primary  {kick, dash, turn, move, catch}
 * and non-primary (all others). It is possible to sent many of non-primary
 * commands at each cycle, whereas the primary commands can be sent only once.
 */
class ServerCommands{
  private:
    /** contains the primary command */
    char mainCommand[MAIN_COMM_SIZE];
    /** contains the type of the primary command */
    Command mainType;
    /** notes if the primary command is set */
    bool mainSet;

    /** contains all the other commands */
    char others[MAX_NO_OTHERS][OTHER_COMM_SIZE];
    /** contains the types of the other commands */
    Command othersTypes[MAX_NO_OTHERS];

    /** index of the last command set */
    int commandsSet;
    /** index of the last command got */
    int commandsGot;

  public:
    ServerCommands();

    /**
     * Sets the primary command. If already set, overwrites it
     * @param msg the command to set
     * @param com the type of the command
     */
    void setMain(const char *msg, Command com)
    { strcpy(mainCommand, msg); mainType = com; mainSet = true; }

    /** @return currently stored primary command. */
    const char*   getMain() { mainSet = false; return mainCommand; }

    /** @return the type of currently stored primary command. */
    Command getMainType()   { return mainType; }

    /** @return true, if primary command is set, false otherwise */
    bool isMainSet() const { return mainSet; }

    /**
     * adds a non-primary command to the queue.
     * @param msg the command to add
     * @param com the type of the command
     */
    void addOther(const char *msg, Command com){
      strcpy(others[commandsSet], msg);
      othersTypes[commandsSet] =  com;
      commandsSet++;
    }

    /**
     * copies the next queued non-primary command in the buffer pointed by the
     * parameter
     * (in the order "first in, first out")
     * @param str the string to copy the command into
     * @return false, if no command was there, true otherwise
     */
    bool getNextOther(char *str);
    Command getOthersType() { return othersTypes[commandsGot-1]; }

    /** Deletes all currently stored commands */
    void deleteAll() { commandsSet = 0; commandsGot = 0; mainSet = false; }

    /** lists all commands */
    void listContent();
};


/**
 * Serves as container for notices about soccer server commands.
 */
class ServerCommandHistory {
  public:
    ServerCommandHistory();
    virtual ~ServerCommandHistory();

    /** \name Accessors */
    //@{
    Command getFst() const { return fst; }
    Command getSnd() const { return snd; }
    Command getTrd() const { return trd; }
    float getFstParam1() const { return fstParam1; }
    float getFstParam2() const { return fstParam2; }
    float getSndParam1() const { return sndParam1; }
    float getSndParam2() const { return sndParam2; }
    const char * getTrdParam1() const { return trdParam1; }

    void setFst(const Command c) { fst = c; }
    void setSnd(const Command c) { snd = c; }
    void setTrd(const Command c) { trd = c; }
    void setFstParam1(const float d) { fstParam1 = d; }
    void setFstParam2(const float d) { fstParam2 = d; }
    void setSndParam1(const float d) { sndParam1 = d; }
    void setSndParam2(const float d) { sndParam2 = d; }
    void setTrdParam1(const char *s)  { strcpy(trdParam1, s); }
    //@}

    /** Deletes all stored data */
    inline void clear();

  void operator=(const ServerCommandHistory &sch){
    fst = sch.fst;
    snd = sch.snd;
    trd = sch.trd;
    fstParam1 = sch.fstParam1;
    fstParam2 = sch.fstParam2;
    sndParam1 = sch.sndParam1;
    sndParam2 = sch.sndParam2;
    strcpy(trdParam1, sch.trdParam1);
  }

    /** output the contents on a string stream */
    friend ostream& operator<< (ostream &os, const ServerCommandHistory &sch){
      os<<"fst: "<<sch.fst<<" fstParam1: "<<sch.fstParam1<<" fstParam2: "<<sch.fstParam2<<endl;
      os<<"snd: "<<sch.snd<<" sndParam1: "<<sch.sndParam1<<" sndParam2: "<<sch.sndParam2<< endl;
      os<<"trd: "<<sch.trd<<" trdParam1: "<<sch.trdParam1<<endl;
      return os;
    }
  private:
    Command fst;  /**< kick, turn, dash, move, coalieCatch */
    Command snd;  /**< turnNeck, changeView, score */
    Command trd;  /**< say */

    float fstParam1; /**< x-coord, power, angle, direction*/
    float fstParam2; /**< y-coord, direction*/

    float sndParam1; /**< ViewMode, turn neck angle */
    float sndParam2; /**< ViewQuality */

    char trdParam1[SAY_MSG_MAX_SIZE]; /**< say message */
};


/**
 * summarizes the command-counters
 */
struct ServerCommandCounters{
  int kicks;
  int dashes;
  int turns;
  int says;
  int turnNecks;
  int catches;
  int moves;
  int changeViews;
  int tackles;
  int pointTos;
  int attentionTos;
  void clear(){
    kicks = dashes = turns = says = turnNecks = catches =
      moves = changeViews = tackles = pointTos = attentionTos = 0;
  }
  int getImportant(){
    return kicks + dashes + turns;
  }
};

void ServerCommandHistory::clear(){
  fst = snd = trd = CMD_NONE;
  fstParam1 = fstParam2 = sndParam1 = sndParam2 = UNDEFINED_FLOAT;
  strcpy(trdParam1, "");
}

// $Source:
#endif // SERVERCOMMANDS_HH
