/******************************************************************************
 * Behaviour.hh
 * Copyright (c) 2003 Virtual Werder
 *
 *
 * implements:
 *   class Behaviour
 *
 * requires:
 *
 *
 * Created on: ??.??.2003 by majunor        
 * Maintainer: tbogon
 * Group:
 * Co-authors: harlekin, cniehaus
 *
 * Changelog:
 *
 * $Id: Behaviour.hh,v 1.21 2004/06/11 11:24:19 harlekin Exp $
 *
 ******************************************************************************/
#ifndef __BEHAVIOUR_H
#define __BEHAVIOUR_H

#include "WorldModel.hh"
#include "SkillPool.hh"
#include "Geometry.hh"
#include "math.h"

#include <map>

/** This class Models the Main Behaviour of the Virtual Werder Agents. Eventual Subclasses may specify different Behaviour or more specific Behaviour.
@todo add all SkillPool Methods here that contain Behaviour or extract the Behaviour from the above mentioned Methods
*/
class Behaviour{
public:
  /** The Constructor initializes the (mostly) private Variables*/
  Behaviour();

  /** Destructor destroying previously instantiated Variables*/
  virtual ~Behaviour();

  /** The standard play On Behaviour calls specific Behaviours*/
  virtual void playOnBehaviour();
  virtual bool acceptPass();

  virtual void reportStatus();

  /** Returns if i am currently at a good positiohn to score a Goal
      @return the Vector where on should shoot if it is a good Point
      or a NULL Pointer if this is a bad point
  */
  //virtual Vector2D isAGoodPointToScore();

  /** Returns if running free is a good choice.
      A Good Choice is the following:
      1. I'm currently intersected from the Player with the Ball
      2. If I run 10m in front, i will be no longer intersected
      @param player the Vector of the Player that should run free
      @return true, if it is a good choice, false otherwise
  */
  bool shouldRunFree(const Vector2D &player);
  virtual bool shouldDribble();
  bool shouldPassTo(const Vector2D &pos);

  virtual bool considerDribbling();
  virtual bool considerCouldPass();
  virtual bool considerShouldPass();

  /** The general Behaviour that tells the Agent what to do in the current Situation, e.g Free Kick, play on*/
  virtual void generalBehaviour();
  
  virtual float getDribbleAngle();
  virtual bool  getDribbleSide();
  virtual int   getDribbleDistance(float angle);


protected:


  /** Handles Penalty Shootout*/
  virtual void doPenalty();
  
  /** Handles the Behaviour if the Agent has the Ball (the Ball is kickable)*/
  virtual void playOnWithBall();

  /** Handles the behaviour if the Agent does not have the Ball (the Ball is not kickable)*/
  virtual void playOnWithoutBall();

  /** Handles the behaviour if the agent doesnt have the ball but a teammate does */
  virtual void assistTeammate();

  /** Returns the potential Pass Candidate
      @return the Pass Candidate
      @param offensive if this is an offensive pass, default value is true
  */
  virtual const Player* getPassCandidate(bool offensive=true);

  /** Returns if stopping the Ball is necessary
      @return true, if the Ball needs to be stopped, false othewise
  */
  virtual bool isStopBallNeeded();

  /** Returns if the Player needs to turn to a Pass Reciever
      @param passCand the PassCandidate
      @return true, if a turn is needed, false otherwise
  */
  virtual bool isTurnToPassCandNeeded(const Player* passCand);

  /** Returns if the currentSeeTime is equal to the currentTime
      @return true if the currentTime is equal to the last See
  */
  virtual bool isLastSeeAtCurrentTime();

 
  /** Performs our kick in. The Player, whose Homeposition is nearest the
   *  kick in Point goes there, the others doPositioning(). The kick in
   *  uses prioritised passing, if no candidate is found he makes a kickaway
   *  (own half) or passes to the mid-attacker (enemy half).
   *  With a high value of wantedRecoveryDelay the time needed for performing
   *  the kick in will be about 0.6 * wantedRecoveryDelay.
   */
  void kickInBehaviour( int wantedRecoveryDelay = 30 );
  
  /** The WorldModel from ehich Data of the Agents "world view" is taken*/
  WorldModel* worldModel;
  /** The Agent itself, containing Data like POsitioning etc.*/
  Myself* myself;
  /** The Ball, quite obvious isn't it :-)*/
  Ball* ball;
  /** The SkillPool */
  SkillPool* skillPool;
  /** Determines which passing should be used*/
  bool usePriorityPassing;
  /** Determines whether a Player is currently running itself free*/
  bool runsFree;
  /** is a Player currently in a Dribbling, then do not stop the Ball*/
  bool currentlyDribbling;

  
   /**
   * Calculates and lets the agent go to the strategic position.
   * This position depends mainly on one (for each teammate different) basic
   * position, the ball position and a few "configuration" parameters. For
   * details see the implementation.
   */
  bool goToStrategicPosition();
  /** avoid the ball if we are not allowed there **/
  void goToStrategicPosition_StayAway(Vector2D &pos);
  Vector2D getStrategicPosition();

  /** summarizes the positioning-behaviour of an field player */  
  virtual void doPositioning();
  
  /**
   * looks for the pass receiver and passes if one was found.
   * @param offensive specifies, if it should be looked for an offensive pass
   * @param endSpeed the speed that the ball should habe at the end of a pass
   * @return true, if the ball was passed, false otherwise
   */
   bool passIfHaveACandidate(bool offensive, float endSpeed);
  
   /**
   * Lets the goalie perform a kick in.
   * @param offensive specifies, if it should be looked for an offensive pass
   * @param minDelay time to waste (in cycles)
   * @param scanDelay time (in cycles) to search the pass receiver
   * @param endSpeed the speed that the ball should habe at the end of a pass
   * @return true if execution ready, false otherwise
   */
   bool performKickIn(bool offensive, int minDelay,
                              int scanDelay, float endSpeed);
                              
   /** Kicks the Ball to a Point near the Center of the Panelty Area of the Opponents Goal*/
   void kickToMidfielder();
   
   /** The positioning of the goalie */
  // Now he uses the derivated doPositioning
   //void goaliePositioning();
   
  bool shouldTryToScoreGoal();
  bool tryToScoreGoal();
  bool tryToScoreGoalWithGoalie();
    bool tryToScoreGoalWithoutGoalie();  
  
  bool tryToForceScoreGoal();

  bool shouldPerformKickIn(const Vector2D &pos, bool &assist);
  void freeKickBehaviour();

  void resetKickIn();
  
  int kickInMinDelay ;
  int kickInScanDelay;
  int kickInState;

  Vector2D passPoint;
  Vector2D passVelocity;
  const Player* passCand;
  bool canPassDirect;
  bool canPassIndirect;
  bool shouldPassDirect;
  bool shouldPassIndirect;

  /** Checks the Playmode and performs the penalty kick */
  bool doPenaltyScore();
  //bool doPenalty();
  
  void lookAround(bool forceBall = false);
  void lookAround_GoalieKickOffPoint(float &angle);


private:

  /** Determines whether the current Player Position is god to score a goal on penalties
      @return true if it is a good point, false otherwise
  */
  bool isAGoodPenaltyShotPoint();
};

#endif

