#ifndef __PLANNERBEHAVIOUR_H
#define __PLANNERBEHAVIOUR_H

#include "WorldModel.hh"
#include "SkillPool.hh"
#include "Geometry.hh"
#include "math.h"

#include <map>

/** This class Models the Main Behaviour of the Virtual Werder Agents. Eventual Subclasses may specify different Behaviour or more specific Behaviour.
@todo add all SkillPool Methods here that contain Behaviour or extract the Behaviour from the above mentioned Methods
*/
class PlannerBehaviour : public Behaviour
{
public:
  /** The Constructor initializes the (mostly) private Variables*/
  PlannerBehaviour();

  /** Destructor destroying previously instantiated Variables*/
   ~PlannerBehaviour();

  /** The standard play On Behaviour calls specific Behaviours*/
   void playOnBehaviour();

  /** Returns if i am currently at a good positiohn to score a Goal
      @return the Vector where on should shoot if it is a good Point
      or a NULL Pointer if this is a bad point
  */
   Vector2D isAGoodPointToScore();

  /** Returns the Pass Candidate after the Priority based Model
      @param offensive is this an offensive Player, will be removed
      @return the Pass Candidate or NULL
  */
   const Player* getPrioritisedPassCandidate(bool offensive=true);

  /** Returns if running free is a good choice.
      A Good Choice is the following:
      1. I'm currently intersected from the Player with the Ball
      2. If I run 10m in front, i will be no longer intersected
      @param player the Vector of the Player that should run free
      @return true, if it is a good choice, false otherwise
  */
  bool shouldRunFree(Vector2D player);

  /** The general Behaviour that tells the Agent what to do in the current Situation, e.g Free Kick, play on*/
  void generalBehaviour();
  
 bool getDribbleSide();


protected:


  /** Handles Penalty Shootout*/
   void doPenalty();
  
  /** Checks if a Pass to the Midfielder is possible
      @param empty Vector2D Pointer. If true is returned, this contains the Coordinates where one should pass to.
      If false is returned this will be a NULL Pointer !
      @return true if the Pass is possible, false otherwise
   */
   bool isPassToMidfielderPossible(Vector2D* midfielderPosition);

  /** Handles the Behaviour if the Agent has the Ball (the Ball is kickable)*/
   void playOnWithBall();

  /** Handles the behaviour if the Agent does not have the Ball (the Ball is not kickable)*/
   void playOnWithoutBall();

  /** Returns the potential Pass Candidate
      @return the Pass Candidate
      @param offensive if this is an offensive pass, default value is true
  */
   const Player* getPassCandidate(bool offensive=true);

  /** Returns if stopping the Ball is necessary
      @return true, if the Ball needs to be stopped, false othewise
  */
   bool isStopBallNeeded();

  /** Returns if the Player needs to turn to a Pass Reciever
      @param passCand the PassCandidate
      @return true, if a turn is needed, false otherwise
  */
   bool isTurnToPassCandNeeded(const Player* passCand);

  /** Returns if the currentSeeTime is equal to the currentTime
      @return true if the currentTime is equal to the last See
  */
   bool isLastSeeAtCurrentTime();

  /** Returns if a Dribbling by a Wing Attacker is possible, e.g. 
      his way is free to the Goal etc.
      @param dribblePOint the POint to which the wing Attacker would 
      dribble if it is possible
      @return true if a dribble is possible, false otherwise
  */
   bool isWingAttackerDribblePossible(Vector2D dribblePoint);

  /** Performs our kick in. The Player, whose Homeposition is nearest the
   *  kick in Point goes there, the others doPositioning(). The kick in
   *  uses prioritised passing, if no candidate is found he makes a kickaway
   *  (own half) or passes to the mid-attacker (enemy half).
   *  With a high value of wantedRecoveryDelay the time needed for performing
   *  the kick in will be about 0.6 * wantedRecoveryDelay.
   */
  void kickInBehaviour( int wantedRecoveryDelay = 30 );
  
//  /** The WorldModel from ehich Data of the Agents "world view" is taken*/
//  WorldModel* worldModel;
//  /** The Agent itself, containing Data like POsitioning etc.*/
//  Myself* myself;
//  /** The Ball, quite obvious isn't it :-)*/
//  Ball* ball;
//  /** Determines which passing should be used*/
//  bool usePriorityPassing;
//  /** Determines whether a Player is currently running itself free*/
//  bool runsFree;
//  /** is a Player currently in a Dribbling, then do not stop the Ball*/
//  bool currentlyDribbling;
  
  void debugHowMuchPlayers();
  
   /**
   * Calculates and lets the agent go to the strategic position.
   * This position depends mainly on one (for each teammate different) basic
   * position, the ball position and a few "configuration" parameters. For
   * details see the implementation.
   */
  bool goToStrategicPosition();

  /** summarizes the positioning-behaviour of an field player */  
  void doPositioning();
  
  /**
   * looks for the pass receiver and passes if one was found.
   * @param offensive specifies, if it should be looked for an offensive pass
   * @param endSpeed the speed that the ball should habe at the end of a pass
   * @return true, if the ball was passed, false otherwise
   */
   bool passIfHaveACandidate(bool offensive, float endSpeed);
  
   /**
   * Lets the goalie perform a kick in.
   * @param offensive specifies, if it should be looked for an offensive pass
   * @param minDelay time to waste (in cycles)
   * @param scanDelay time (in cycles) to search the pass receiver
   * @param endSpeed the speed that the ball should habe at the end of a pass
   * @return true if execution ready, false otherwise
   */
   bool performKickIn(bool offensive, int minDelay,
                              int scanDelay, float endSpeed);
                              
   /** Kicks the Ball to a Point near the Center of the Panelty Area of the Opponents Goal*/
   void kickToMidfielder();
   
   /** The positioning of the goalie */
   void goaliePositioning();
   

  /**
   * Run to a position where you can be used as pass candidate.
   * @return true if a TM has the ball.
   */
   bool offerAsPassCandidate();
   
   void resetKickIn();
  
  int kickInMinDelay ;
  int kickInScanDelay;
  int kickInState;

private:
  // --> Constants, these should be in Configuration Files in the near future
  /** The number of the Middle Attacker*/
  static const int UNUM_MIDDLEATTACKER = 10;
  
  /** Determines whether the current Player Position is god to score a goal on penalties
      @return true if it is a good point, false otherwise
  */
  bool isAGoodPenaltyShotPoint();
};

#endif
