/******************************************************************************
 * BodyInfo.hh
 * Copyright (c) 2001 Virtual Werder
 *
 * implements:
 *   class BodyInfo
 *
 * requires:
 *
 * Created on: 29.05.2001 by mag
 * Maintainer: harlekin
 * Group:
 * Co-authors: wenns
 *
 * Changelog:
 *
 *****************************************************************************/

#ifndef BODYINFO_HH
#define BODYINFO_HH

#include <iostream>
#include "common.hh"
#include "Geometry.hh"

/**
 * Contains the agent body-relevant information. All this information is
 * gained from a sense body message.
 */
class BodyInfo {
  public:
    /** Default constructor, the time and the counter are initialized with 0 */
    BodyInfo();
    virtual ~BodyInfo();

    /** \name Accessors */
    //@{
    ViewQuality getViewQuality () const {return viewQuality;}
    ViewWidth getViewWidth ()     const {return viewWidth;}
    float getStamina ()           const {return stamina;}
    float getEffort ()            const {return effort;}
    float getHeadAngle ()         const {return headAngle;}
    int getKickCount ()           const {return kickCount;}
    int getDashCount ()           const {return dashCount;}
    int getTurnCount ()           const {return turnCount;}
    int getSayCount ()            const {return sayCount;}
    int getTurnNeckCount ()       const {return turnNeckCount;}
    int getCatchCount ()          const {return catchCount;}
    int getMoveCount ()           const {return moveCount;}
    int getChangeViewCount ()     const {return changeViewCount;}
  int getImportantCounts() const {return kickCount + dashCount + turnCount; }

    int getTackleExpires ()         const {return tackleExpires;}
    int getTackleCount ()           const {return tackleCount;}

    bool      isFocusSet ()         const {return focusSet;}
    FieldSide getFocusTargetSide () const {return focusTargetSide;}
    int       getFocusTargetUnum () const {return focusTargetUnum;}
    int       getFocusCount ()      const {return focusCount;}

    bool      isArmMovable ()       const {return armMovable;}
    int       getArmExpires ()      const {return armExpires;}
    Vector2D  getArmTarget ()       const {return armTarget;}
    int       getArmCount ()        const {return armCount;}


    void setAll( const ViewQuality newViewQuality,
		 const ViewWidth newViewWidth, const float newStamina,
		 const float newEffort, const float newHeadAngle,
		 const int newKickCount, const int newDashCount,
		 const int newTurnCount, const int newSayCount,
		 const int newTurnNeckCount, int newCatchCount,
		 const int newMoveCount, const int newChangeViewCount)
    {
      viewQuality     = newViewQuality;
      viewWidth       = newViewWidth;
      stamina         = newStamina;
      effort          = newEffort;
      headAngle       = newHeadAngle;
      kickCount       = newKickCount;
      dashCount       = newDashCount;
      turnCount       = newTurnCount;
      sayCount        = newSayCount;
      turnNeckCount   = newTurnNeckCount;
      catchCount      = newCatchCount;
      moveCount       = newMoveCount;
      changeViewCount = newChangeViewCount;
    }


    void setTackleInfo( const int expires, const int count){
      tackleExpires = expires;
      tackleCount   = count;
    }

    void setFocusInfo( const bool set, const FieldSide side, const int unum,
                       const int count){
      focusSet        = set;
      focusTargetSide = side;
      focusTargetUnum = unum;
      focusCount      = count;
    }

    void setArmInfo( const bool movable, const int expires,
                     const Vector2D &target, const int count ){
      armMovable = movable;
      armExpires = expires;
      armTarget  = target;
      armCount   = count;
    }
    //@}

    /** output on a string stream */
    friend ostream& operator<< (ostream &os, const BodyInfo &bi){
      os << " viewQuality: " << bi.viewQuality
        << " viewWidth: " << bi.viewWidth << " stamina: " << bi.stamina
        << " effort: " << bi.effort
        << " headAngle: " << bi.headAngle
        << " kickCount: " << bi.kickCount << " dashCount: " << bi.dashCount
        << " turnCount: " << bi.turnCount << " sayCount: " << bi.sayCount
        << " turnNeckCount: " << bi.turnNeckCount
        << " catchCount: " << bi.catchCount
        << " moveCount: " << bi.moveCount
        << " changeViewCount: " << bi.changeViewCount
        << endl;
      return os;
    }

  private:
    /** The soccer server cycle the information comes from */
    ViewQuality viewQuality;   /**< HIGH or LOW*/
    ViewWidth   viewWidth;     /**< WIDE, NORMAL or NARROW*/
    float stamina;             /**< max 4000 by default */
    float effort;              /**< [0.6 .. 1) by default */
    float headAngle;           /**< relative to the body */


    /**
     * \name Command counter
     *
     * The counter serve primarily as a debugging goody. The server increments
     * a counter, if an according command was received, and sends all the
     * counter values to the client with the next sense body message.
     */
    //@{
    int kickCount;
    int dashCount;
    int turnCount;
    int sayCount;
    int turnNeckCount;
    int catchCount;
    int moveCount;
    int changeViewCount;
    //@}

    /** \name tackle information*/
    //@{
    int tackleExpires;   /**< amount of cycles until the player can move
                              after a tackle, 0 indicates the player isnt
                              tackling */
    int tackleCount;     /**< number of tackles performed by player */
    //@}

    /** \name focus information*/
    //@{
    bool      focusSet;        /**< true, if the focus is currently set */
    FieldSide focusTargetSide; /**< team affiliation of the focused player */
    int       focusTargetUnum; /**< unum of the focused player */
    int       focusCount;      /**< number of attentionto actions performed
                                    by player */
    //@}

    /** \name arm information*/
    //@{
    int       armMovable;     /**< number of cycles, till the arm is movable
                                   0 means the arm is movable now */
    int       armExpires;     /**< numver of cycles till the arm stops
                                   pointing 0 means the arm isnt pointing */
    Vector2D  armTarget;      /**< the absolute field position the arm
                                   is pointing */
    int       armCount;       /**< number of pointto actions performed by
                                   player */
    //@}
};

#endif //BODYINFO_HH
