/******************************************************************************
 * Field.cc
 * Copyright (c) 2001 Virtual Werder
 *
 * implements:
 *   class Field
 *
 * requires:
 *
 *
 * Created on: 22.08.2001 by wenns
 * Maintainer: wenns
 * Group:
 * Co-authors:
 *
 * Changelog:
 *
 ******************************************************************************/

#include "Field.hh"
#include "Configuration.hh"
#include "debug.hh"
#include "common.hh"


const float Field::GOAL_WIDTH              =  14.02;
const float Field::LENGTH                  = 105.0;
const float Field::LENGTH_LINE_BEHIND_GOAL = 110.0;
const float Field::WIDTH                   =  68.0;
const float Field::PENALTY_AREA_LENGTH     =  16.5;
const float Field::PENALTY_AREA_WIDTH      =  40.32;
const float Field::GOAL_AREA_LENGTH        =   5.5;
const float Field::GOAL_AREA_WIDTH         =  18.32;
const float Field::KICKOFF_CIRCLE_RADIUS   =   9.15;
const float Field::PITCH_MARGIN            =   5.0;
const float Field::PENALTY_SHOT_DIST       =  11.0;
const float Field::GOAL_POST_RADIUS        =   0.06;
const int   Field::NUMBER_LINES            =   4;
const int   Field::NUMBER_FLAGS            =  55;


//lines
const float Field::OWN_GOAL_OUTLINE            = -Field::LENGTH / 2;
const float Field::OPP_GOAL_OUTLINE            =  Field::LENGTH / 2;
const float Field::R_SIDE_OUTLINE              =  Field::WIDTH  / 2;
const float Field::L_SIDE_OUTLINE              = -Field::WIDTH  / 2;
const float Field::OWN_LINE_BEHIND_GOALOUTLINE = -Field::LENGTH_LINE_BEHIND_GOAL / 2;
const float Field::OPP_LINE_BEHIND_GOALOUTLINE =  Field::LENGTH_LINE_BEHIND_GOAL / 2;


//field areas
const Rectangle Field::FIELD =
  Rectangle(Vector2D(OWN_GOAL_OUTLINE, L_SIDE_OUTLINE),
	    Vector2D(OPP_GOAL_OUTLINE, R_SIDE_OUTLINE));
const Rectangle Field::OWN_SIDE_AREA =
  Rectangle(Vector2D(OWN_GOAL_OUTLINE, -WIDTH / 2), Vector2D(0, WIDTH / 2));
const Rectangle Field::OPP_SIDE_AREA =
  Rectangle(Vector2D(0, -WIDTH / 2), Vector2D(OPP_GOAL_OUTLINE, WIDTH / 2));
const Rectangle Field::OWN_PENALTY_AREA =
  Rectangle(Vector2D(OWN_GOAL_OUTLINE, - PENALTY_AREA_WIDTH/2),
            Vector2D(OWN_GOAL_OUTLINE  + PENALTY_AREA_LENGTH, PENALTY_AREA_WIDTH/2));
const Rectangle Field::OPP_PENALTY_AREA =
  Rectangle(Vector2D(OPP_GOAL_OUTLINE, - PENALTY_AREA_WIDTH/2),
            Vector2D(OPP_GOAL_OUTLINE  - PENALTY_AREA_LENGTH, PENALTY_AREA_WIDTH/2));
const Rectangle Field::OWN_GOAL_AREA =
  Rectangle(Vector2D(OWN_GOAL_OUTLINE, - GOAL_AREA_WIDTH/2),
            Vector2D(OWN_GOAL_OUTLINE  + GOAL_AREA_LENGTH, GOAL_AREA_WIDTH/2));
const Rectangle Field::OPP_GOAL_AREA =
  Rectangle(Vector2D(OPP_GOAL_OUTLINE, - GOAL_AREA_WIDTH/2),
            Vector2D(OPP_GOAL_OUTLINE  - GOAL_AREA_LENGTH, GOAL_AREA_WIDTH/2));
const Rectangle Field::OWN_PENALTY_TOLERANCE_AREA =
  Rectangle(Vector2D(OWN_GOAL_OUTLINE - 2, - PENALTY_AREA_WIDTH/2),
            Vector2D(OWN_GOAL_OUTLINE  + PENALTY_AREA_LENGTH, PENALTY_AREA_WIDTH/2));

//penalty and kick off related
const Vector2D Field::OWN_PENALTY_POINT = 
  Vector2D(OWN_GOAL_OUTLINE + PENALTY_SHOT_DIST, 0);
const Vector2D Field::OPP_PENALTY_POINT = 
  Vector2D(OPP_GOAL_OUTLINE - PENALTY_SHOT_DIST, 0);
const Circle Field::KICK_OFF_CIRCLE = 
  Circle(Vector2D(0, 0), KICKOFF_CIRCLE_RADIUS);


//goal posts related
const Vector2D Field::OWN_GOAL_POST_TOP_POS =
  Vector2D(-(LENGTH / 2 - GOAL_POST_RADIUS), -(GOAL_WIDTH / 2 + GOAL_POST_RADIUS));
const Vector2D Field::OWN_GOAL_POST_BOTTOM_POS =
  Vector2D(-(LENGTH / 2 - GOAL_POST_RADIUS), (GOAL_WIDTH / 2 + GOAL_POST_RADIUS));
const Vector2D Field::OPP_GOAL_POST_TOP_POS =
  Vector2D((LENGTH / 2 - GOAL_POST_RADIUS), -(GOAL_WIDTH / 2 + GOAL_POST_RADIUS));
const Vector2D Field::OPP_GOAL_POST_BOTTOM_POS =
  Vector2D((LENGTH / 2 - GOAL_POST_RADIUS), (GOAL_WIDTH / 2 + GOAL_POST_RADIUS));
const Circle Field::OWN_GOAL_POST_TOP = 
  Circle(Field::OWN_GOAL_POST_TOP_POS, Field::GOAL_POST_RADIUS);
const Circle Field::OWN_GOAL_POST_BOTTOM =
  Circle(Field::OWN_GOAL_POST_BOTTOM_POS, Field::GOAL_POST_RADIUS);
const Circle Field::OPP_GOAL_POST_TOP =
  Circle(Field::OPP_GOAL_POST_TOP_POS, Field::GOAL_POST_RADIUS);
const Circle Field::OPP_GOAL_POST_BOTTOM =
  Circle(Field::OPP_GOAL_POST_BOTTOM_POS, Field::GOAL_POST_RADIUS);


//goalie related 
const Circle Field::GOALIE_POSITIONING_CIRCLE =
      Circle(Vector2D(OWN_GOAL_OUTLINE, 0), GOAL_WIDTH / 2);
const Circle Field::GET_CATCH_BALL_CIRCLE =
      Circle(Vector2D(OWN_GOAL_OUTLINE, 0), (GOAL_WIDTH / 3) * 2);
const Vector2D Field::GOAL_CENTER = Vector2D(OWN_GOAL_OUTLINE, 0);
const Vector2D Field::CENTER_BEHIND_GOALOUTLINE   = Vector2D(LENGTH_LINE_BEHIND_GOAL, 0);
const Vector2D Field::GOALIE_KICK_OFF_POINT_LEFT  = Vector2D(-40.0, -12.0);
const Vector2D Field::GOALIE_KICK_OFF_POINT_RIGHT = Vector2D(-40.0,  12.0);


//the rectangles are build with the according kick-off-point as the middle of
//the lower horizontal line and y-offset to the right and to the left and a
//point bounding the rectangle above
const Rectangle Field::GOALIE_KICK_OFF_RECT_LEFT =
Rectangle(Vector2D(Field::GOALIE_KICK_OFF_POINT_LEFT.getX(),         // = -38
		   Field::GOALIE_KICK_OFF_POINT_LEFT.getY() + 17.0), // =   5
	  Vector2D(Field::GOALIE_KICK_OFF_POINT_LEFT.getX() + 20.0,  // = -18
		   Field::GOALIE_KICK_OFF_POINT_LEFT.getY() - 13.0));// = -25

const Rectangle Field::GOALIE_KICK_OFF_RECT_RIGHT =
Rectangle(Vector2D(Field::GOALIE_KICK_OFF_POINT_RIGHT.getX(),         // = -38
		   Field::GOALIE_KICK_OFF_POINT_RIGHT.getY() + 13.0), // =  25
	  Vector2D(Field::GOALIE_KICK_OFF_POINT_RIGHT.getX() + 20.0,  // = -18
		   Field::GOALIE_KICK_OFF_POINT_RIGHT.getY() - 17.0));// = - 5


Field::Field(){
  // Those are important. Others can be ignored.
  markerConfidence[GOAL_L] = Time();
  markerConfidence[GOAL_R] = Time();

  //goals
  markerPositions[GOAL_L]  = *(new Vector2D(-LENGTH/2, 0));
  markerPositions[GOAL_R]  = *(new Vector2D( LENGTH/2, 0));

  //center flags
  markerPositions[FLAG_C]  = *(new Vector2D(0, 0));
  markerPositions[FLAG_CT] = *(new Vector2D(0, -WIDTH/2));
  markerPositions[FLAG_CB] = *(new Vector2D(0,  WIDTH/2));

  //corner flags
  markerPositions[FLAG_LT] = *(new Vector2D(-LENGTH/2, -WIDTH/2));
  markerPositions[FLAG_LB] = *(new Vector2D(-LENGTH/2,  WIDTH/2));
  markerPositions[FLAG_RT] = *(new Vector2D( LENGTH/2, -WIDTH/2));
  markerPositions[FLAG_RB] = *(new Vector2D( LENGTH/2,  WIDTH/2));

   //penalty area flags
  markerPositions[FLAG_PLT] =
    *(new Vector2D(-LENGTH/2 + PENALTY_AREA_LENGTH, -PENALTY_AREA_WIDTH/2));
  markerPositions[FLAG_PLC] =
    *(new Vector2D(-LENGTH/2 + PENALTY_AREA_LENGTH, 0));
  markerPositions[FLAG_PLB] =
    *(new Vector2D(-LENGTH/2 + PENALTY_AREA_LENGTH, PENALTY_AREA_WIDTH/2));
  markerPositions[FLAG_PRT] =
    *(new Vector2D(LENGTH/2 - PENALTY_AREA_LENGTH, -PENALTY_AREA_WIDTH/2));
  markerPositions[FLAG_PRC] =
    *(new Vector2D(LENGTH/2 - PENALTY_AREA_LENGTH, 0));
  markerPositions[FLAG_PRB] =
    *(new Vector2D(LENGTH/2 - PENALTY_AREA_LENGTH, PENALTY_AREA_WIDTH/2));

  //jambs
  markerPositions[FLAG_GLT] = *(new Vector2D(-LENGTH/2, -GOAL_WIDTH / 2 ));
  markerPositions[FLAG_GLB] = *(new Vector2D(-LENGTH/2,  GOAL_WIDTH / 2 ));
  markerPositions[FLAG_GRT] = *(new Vector2D( LENGTH/2, -GOAL_WIDTH / 2 ));
  markerPositions[FLAG_GRB] = *(new Vector2D( LENGTH/2,  GOAL_WIDTH / 2 ));

  //top flags
  markerPositions[FLAG_TL50] = *(new Vector2D(-50, -(WIDTH/2 + 5 )));
  markerPositions[FLAG_TL40] = *(new Vector2D(-40, -(WIDTH/2 + 5 )));
  markerPositions[FLAG_TL30] = *(new Vector2D(-30, -(WIDTH/2 + 5 )));
  markerPositions[FLAG_TL20] = *(new Vector2D(-20, -(WIDTH/2 + 5 )));
  markerPositions[FLAG_TL10] = *(new Vector2D(-10, -(WIDTH/2 + 5 )));
  markerPositions[FLAG_T0]   = *(new Vector2D(  0, -(WIDTH/2 + 5 )));
  markerPositions[FLAG_TR10] = *(new Vector2D( 10, -(WIDTH/2 + 5 )));
  markerPositions[FLAG_TR20] = *(new Vector2D( 20, -(WIDTH/2 + 5 )));
  markerPositions[FLAG_TR30] = *(new Vector2D( 30, -(WIDTH/2 + 5 )));
  markerPositions[FLAG_TR40] = *(new Vector2D( 40, -(WIDTH/2 + 5 )));
  markerPositions[FLAG_TR50] = *(new Vector2D( 50, -(WIDTH/2 + 5 )));

  //bottom flags
  markerPositions[FLAG_BL50] = *(new Vector2D(-50, (WIDTH/2 + 5 )));
  markerPositions[FLAG_BL40] = *(new Vector2D(-40, (WIDTH/2 + 5 )));
  markerPositions[FLAG_BL30] = *(new Vector2D(-30, (WIDTH/2 + 5 )));
  markerPositions[FLAG_BL20] = *(new Vector2D(-20, (WIDTH/2 + 5 )));
  markerPositions[FLAG_BL10] = *(new Vector2D(-10, (WIDTH/2 + 5 )));
  markerPositions[FLAG_B0]   = *(new Vector2D(  0, (WIDTH/2 + 5 )));
  markerPositions[FLAG_BR10] = *(new Vector2D( 10, (WIDTH/2 + 5 )));
  markerPositions[FLAG_BR20] = *(new Vector2D( 20, (WIDTH/2 + 5 )));
  markerPositions[FLAG_BR30] = *(new Vector2D( 30, (WIDTH/2 + 5 )));
  markerPositions[FLAG_BR40] = *(new Vector2D( 40, (WIDTH/2 + 5 )));
  markerPositions[FLAG_BR50] = *(new Vector2D( 50, (WIDTH/2 + 5 )));

  //left flags
  markerPositions[FLAG_LT30] = *(new Vector2D(-(LENGTH/2 + 5), -30));
  markerPositions[FLAG_LT20] = *(new Vector2D(-(LENGTH/2 + 5), -20));
  markerPositions[FLAG_LT10] = *(new Vector2D(-(LENGTH/2 + 5), -10));
  markerPositions[FLAG_L0]   = *(new Vector2D(-(LENGTH/2 + 5),   0));
  markerPositions[FLAG_LB10] = *(new Vector2D(-(LENGTH/2 + 5),  10));
  markerPositions[FLAG_LB20] = *(new Vector2D(-(LENGTH/2 + 5),  20));
  markerPositions[FLAG_LB30] = *(new Vector2D(-(LENGTH/2 + 5),  30));

  //right flags
  markerPositions[FLAG_RT30] = *(new Vector2D((LENGTH/2 + 5), -30));
  markerPositions[FLAG_RT20] = *(new Vector2D((LENGTH/2 + 5), -20));
  markerPositions[FLAG_RT10] = *(new Vector2D((LENGTH/2 + 5), -10));
  markerPositions[FLAG_R0]   = *(new Vector2D((LENGTH/2 + 5),   0));
  markerPositions[FLAG_RB10] = *(new Vector2D((LENGTH/2 + 5),  10));
  markerPositions[FLAG_RB20] = *(new Vector2D((LENGTH/2 + 5),  20));
  markerPositions[FLAG_RB30] = *(new Vector2D((LENGTH/2 + 5),  30));
}

Field::~Field(){
  delete &markerPositions;
  delete &markerConfidence;
}

float Field::computeFaceDirection(const vector <Sense::Line> &lines,
                                  FieldSide side){
  //the statements "distance" and "direction" relating to a field side line
  //are to interpret as follows:
  //"distance" is the distance to the intersection of the "sight line"
  // (bisecting line of the see cone) and the aforesaid line.
  //"direction" is an angle between the sight line and the aforesaid line

  // Set to BIG_FLOAT to avoid problems when no angle can be computed
  float angle = BIG_FLOAT;  //face direction
  int lineId = -1;

  switch ( lines.size() ) {
    case 0:
      cout << "Cannot compute face dir. Number of seen lines is 0." << endl;
      return angle;      
    case 1:       //the position of the agent is inside the field
      angle =  lines[0].getRelPos().getDirection();
      lineId = lines[0].getId();
      break;
    default:
      //the only way to see two lines is to be outside of the field and look
      //in the direction of the field, so the agent is outside.
      //the second line is seen from the innerside, so we can calculate the
      //angle as if we were inside the field
      //the second line is the one with the greater distance, so choose
      //accordingly
      int i;
      float length1 = lines[0].getRelPos().getLength();
      float length2 = lines[1].getRelPos().getLength();

      if( length1 > length2 )
        i = 0;
      else
        i = 1;
      angle  = lines[i].getRelPos().getDirection();
      lineId = lines[i].getId();
      break;
  }

  //depending on which line we see, modify the cutting angle sent by the
  //server in order to calculate the absolute face direction of the agent
  switch ( lineId ) {
    case L_Top:
      if ( angle >= 0 ) angle = -angle;
      else              angle = -180 - angle;
      break;
    case L_Bottom:
      if ( angle > 0 )  angle = 180 - angle;
      else              angle = -angle;
      break;
    case L_Left:
      if ( angle > 0 )      angle = -90 - angle;
      else if ( angle < 0 ) angle = 90 - angle;
      else{
        cout << "ERROR: Couldn't compute face dir. Special case 1 is not\
          treated." << endl;
	angle = BIG_FLOAT;
      }
      break;
    case L_Right:
      if ( angle > 0 )       angle = 90 - angle;
      else if ( angle < 0 )  angle = -90 - angle;
      else{
        cout << "ERROR: Couldn't compute face dir. Special case 2 is not\
          treated." << endl;
	angle = BIG_FLOAT;
      }
      break;
    default:
      cout << "ERROR: Unknown line id in computeFaceDir" << endl;
      angle = BIG_FLOAT;
      break;
  }

  if( side == FS_Right && angle != BIG_FLOAT)
    angle = normalizeTo180Deg(angle + 180);

  return angle;
}


Vector2D Field::computePosFromFlag(const Sense::Flag &flag,
                                   float faceDir, FieldSide side){
  //the absolute position of the agent on the soccer server field is
  //the difference of the known absolute position of the flag and the position
  //of the flag relative to the agent

  //Since the absolute flag positions are stored from the perspective of the
  //left team, a negation of the positions are needed in the case we are
  //playing right.

  Vector2D flagPos, myPos;

  flagPos = flag.getRelPos();
  // rotate changes object. No need to save it any other way
  flagPos.rotate(faceDir);

  if( side == FS_Left )
    myPos =   getMarkerPosition(flag.getId()) - flagPos;
  if( side == FS_Right )
    myPos = - getMarkerPosition(flag.getId()) - flagPos;

  return myPos;
}

