/**************************************************************************************
 * Field.hh
 * Copyright (c) 2001 Virtual Werder
 *
 * implements:
 *   class Field
 *
 * requires:
 *
 *
 * Created on: 22.08.2001 by wenns
 * Maintainer: wenns
 * Group:
 * Co-authors:
 *
 * Changelog: the hash_map replaced by map because of hash_map isnt part of the
 * c++ standard
 *
 *************************************************************************************/

#ifndef FIELD_HH
#define FIELD_HH

//System includes
#include <map>
#include <vector>

//Project includes
#include "Geometry.hh"
#include "SenseObjects.hh"
#include "Time.hh"

/**
 * This class contains all information about the robocup soccer simulation
 * field. Among others, it contains the absolute positions of the flags and
 * helper functions to calculate the own position
 */
class Field{
  private:
    /** stores pairs of marker-id's and whose absolute positions on the field */
  map<int, Vector2D> markerPositions;
  map<int, Time>     markerConfidence;
  
  public:
    /** the width of the goal */
    static const float Field::GOAL_WIDTH;
    /** the length of the soccer field */
    static const float Field::LENGTH;
    /** the length from the endline of the goal behind the outline*/
    static const float Field::LENGTH_LINE_BEHIND_GOAL;
    /** the width of the soccer field */
    static const float Field::WIDTH;
    /** the length of the penalty area */
    static const float Field::PENALTY_AREA_LENGTH;
    /** the width of the penalty area */
    static const float Field::PENALTY_AREA_WIDTH;
    /** the length of the goal area */
    static const float Field::GOAL_AREA_LENGTH;
    /** the width of the goal area */
    static const float Field::GOAL_AREA_WIDTH;
    /** the radius of the middle circle */
    static const float Field::KICKOFF_CIRCLE_RADIUS;
    /** the margin of the pitch */ 
    static const float Field::PITCH_MARGIN;
    /** the distance from penalty point to goal */
    static const float Field::PENALTY_SHOT_DIST;
    /** the radius of the goal posts */
    static const float Field::GOAL_POST_RADIUS; 
    /** the number of lines on the field */
    static const int Field::NUMBER_LINES;            
    /** the number of flags */
    static const int Field::NUMBER_FLAGS;           


    /** the x-coordinate of the own goal-out-line */
    static const float Field::OWN_GOAL_OUTLINE;
    /** the x-coordinate of the opponent goal-out-line */
    static const float Field::OPP_GOAL_OUTLINE;
    /** the y-coordinate of the right side-out-line */
    static const float Field::R_SIDE_OUTLINE;
    /** the y-coordinate of the left side-out-line */
    static const float Field::L_SIDE_OUTLINE;
    /** the x-coordinate of the own endline of the goal behind the outline*/
    static const float Field::OWN_LINE_BEHIND_GOALOUTLINE;
    /** the x-coordinate of the opponent endline of the goal behind the outline */  
    static const float Field::OPP_LINE_BEHIND_GOALOUTLINE;
    
    
    /** a rectangle that represents the field */
    static const Rectangle Field::FIELD;
    /** a rectangle that represents own half of the field*/
    static const Rectangle Field::OWN_SIDE_AREA;
    /** a rectangle that represents the opponent half of the field*/
    static const Rectangle Field::OPP_SIDE_AREA;
    /** a rectangle that represents the own penalty area */
    static const Rectangle Field::OWN_PENALTY_AREA;
    /** a rectangle that represents the opponent penalty area */
    static const Rectangle Field::OPP_PENALTY_AREA;
    /** a rectangle that represents the own goal area */
    static const Rectangle Field::OWN_GOAL_AREA;
    /** a rectangle that represents the opponent goal area */
    static const Rectangle Field::OPP_GOAL_AREA;
    // a rectangle that represents the own goal area with
    // a little tolerance behind the goal - for goalie use
    static const Rectangle Field::OWN_PENALTY_TOLERANCE_AREA;

    /** the own penalty point */
    static const Vector2D Field::OWN_PENALTY_POINT;
    /** the opponent penalty point */
    static const Vector2D Field::OPP_PENALTY_POINT;
    /** the kick off circle */
    static const Circle Field::KICK_OFF_CIRCLE;


    /** the position of the own top goal post */
    static const Vector2D Field::OWN_GOAL_POST_TOP_POS;
    /** the position of the own bottom goal post */
    static const Vector2D Field::OWN_GOAL_POST_BOTTOM_POS;
    /** the position of the opponent top goal post */
    static const Vector2D Field::OPP_GOAL_POST_TOP_POS;
    /** the position of the opponent bottom goal post */
    static const Vector2D Field::OPP_GOAL_POST_BOTTOM_POS;
    /** the own top goal post */
    static const Circle Field::OWN_GOAL_POST_TOP;
    /** the own bottom goal post */
    static const Circle Field::OWN_GOAL_POST_BOTTOM;
    /** the opponent top goal post */
    static const Circle Field::OPP_GOAL_POST_TOP;
    /** the opponent bottom goal post */
    static const Circle Field::OPP_GOAL_POST_BOTTOM;
  

    /** goalie related */
    /** semicircle in front of the own goal for the goalie positioning */
    static const Circle Field::GOALIE_POSITIONING_CIRCLE;
    /** semicircle in front of goal where goalie shall always try catch */
    static const Circle Field::GET_CATCH_BALL_CIRCLE;
    /** the center of the own goal */
    static const Vector2D Field::GOAL_CENTER;
    /** the center of the own goal-endline behind the outline*/
    static const Vector2D Field::CENTER_BEHIND_GOALOUTLINE;
    /** the left of two possible points to kick off after a successful catch */
    static const Vector2D Field::GOALIE_KICK_OFF_POINT_LEFT;
    /** the right of two possible points to kick off after a successful catch */    
    static const Vector2D Field::GOALIE_KICK_OFF_POINT_RIGHT;


    /**
     * the left rectangle in which the goalie looks for teammates/opponents
     * in order to decide which kick-off-point is to take
     */
    static const Rectangle Field::GOALIE_KICK_OFF_RECT_LEFT;
    /**
     * the right rectangle in which the goalie looks for teammates/opponents
     * in order to decide which kick-off-point is to take
     */
    static const Rectangle Field::GOALIE_KICK_OFF_RECT_RIGHT;


    Field();
    virtual ~Field();

    /**
     * Returns the absolute position of a marker
     * @param markerID the ID of the marker whose position should be returned
     * @return the position of the marker
     */
    Vector2D getMarkerPosition(const int markerId){
      return markerPositions[markerId];
    }

  float getMarkerConfidence(const int markerId, const Time *now){
    float t = (float)(*now - markerConfidence[markerId]).getCycles();
    return min(max(0.0f, 1.0f - (t/100.0f)), 1.0f);
  }
  
  void setMarkerConfidence(const int markerId, Time *t){
    markerConfidence[markerId] = *t;
  }

    /**
     * calculates the own face direction
     * @param lines parsed information about the seen lines
     * @param side the own field side
     * @return the face direction in degree [0..355]
     */
    float computeFaceDirection(const vector <Sense::Line> &lines, FieldSide side);


    /**
     * calculates the own position on the field using the relative position
     * to a seen flag, the absolute position of the flag and the face
     * direction
     * @param flag parsed information about a seen flag
     * @param faceDir face direction ot the agent
     * @return absolute position of the agent on the field
     */
    Vector2D computePosFromFlag(const Sense::Flag &flag,
                                float faceDir, FieldSide side);
};

#endif //FIELD_HH
