/******************************************************************************
 * Player.cc
 * Copyright (c) 2001 Virtual Werder
 *
 * part of simple agent
 *
 * implements:
 *   class Player
 *
 * requires:
 *   <none>
 *
 * Created on: 22.08.2001 by wenns
 * Maintainer: wenns
 * Group:
 * Co-authors:
 *
 * Changelog:
 *
 ******************************************************************************/

#include "Geometry.hh"
#include "Configuration.hh"
#include "debug.hh"
#include "Player.hh"


Player::Player() : MovingObject(),
    bodyDir(UNDEFINED_FLOAT), 
    headDir(UNDEFINED_FLOAT),
    uniformNumber(UNDEFINED_INT), 
    goalKeeper(false), 
    opponent(false),
    close(false), 
    tackling(false), 
    pointDir(UNDEFINED_FLOAT),
    type(0), 
    stamina(UNDEFINED_FLOAT),
    effort(UNDEFINED_FLOAT), 
    recovery(UNDEFINED_FLOAT)
{
}

Player::Player(const Vector2D &pos, const Vector2D &vel, const float bDir,
               const float hDir, const int unum, const bool goalie,
               const bool opp, const bool cls, const bool tckl,
               const float pd)
  : MovingObject(pos, vel),
    bodyDir(bDir), headDir(hDir),
    uniformNumber(unum), goalKeeper(goalie), opponent(opp), close(cls),
    tackling(tckl), pointDir(pd), type(0), stamina(UNDEFINED_FLOAT),
    effort(UNDEFINED_FLOAT), recovery(UNDEFINED_FLOAT)
{
}

Player::~Player(){
}


bool Player::operator == (const Player &player) const {
  return (uniformNumber==player.getUniformNumber()
	  && opponent==player.isOpponent());
}


Vector2D Player::predictedPosition(int cycles, float dashPower,
                                   float turnAngle) const {
  if(ASSERT_DEBUG) assert(cycles >= 0);
  if(ASSERT_DEBUG) assert(dashPower <= 100 && dashPower >= -100);
  if(ASSERT_DEBUG) assert(turnAngle <= 180 && turnAngle >= -180);

  if (position.isInvalid() || velocity.isInvalid()){
    if (position.isInvalid()){
      DEBUG1("Player::predictedPosition(): Can't estimate future if don't know present");
    }
    return position;
  }
  float playerSpeedMax = Configuration::getMaximumPlayerSpeed(type);
  float playerDecay    = Configuration::getPlayerDecay(type);
  float dashPowerRate  = Configuration::getDashPowerRate(type);
  float inertiaMoment  = Configuration::getInertiaMoment(type);


  float currTurnAng = turnAngle;
  float currBodyDir = bodyDir;
  float effPower, thisTurn;
  static Vector2D currPosition;
  currPosition = position;
  static Vector2D currVelocity;
  currVelocity = velocity;
  static Vector2D acceleration;

  for (int i = 0; i < cycles; i++){
    if (currTurnAng != 0.0) {
      //if a turn necessary ...
      float actualTurnAngle = Configuration::maxmoment /
	(1.0 + inertiaMoment * currVelocity.getLength() );
      
      thisTurn =  min(max(-actualTurnAngle, currTurnAng), actualTurnAngle);
      
      //simply add them to the bodyDir
      currBodyDir = normalizeTo180Deg(currBodyDir + thisTurn);
      currTurnAng -= thisTurn;
      currVelocity.rotate(thisTurn);
    }else{
      effPower = dashPower * dashPowerRate;
      acceleration.setVector2D(effPower, currBodyDir, POLAR);
      currVelocity += acceleration;

      //if the calculated velocity exceeds the maximum
      if ( currVelocity.getLength() > playerSpeedMax )
        //reduce it again
        currVelocity *= ( playerSpeedMax / currVelocity.getLength() );
    }
    currPosition += currVelocity;
    //do decelerate
    currVelocity *= playerDecay;
  }
  return currPosition;
}

float Player::correctDashPowerForStamina(const float dashPower,
					 const float stamina) const 
{
  float newPower;
  int effortDecThreshold = (int)(Configuration::getStaminaMax(type) * Configuration::effortDecThr);

  if (dashPower >= 0) {
    //running forward
    newPower = min((float)dashPower, (float)(stamina - (effortDecThreshold + Player::TIRED_BUFFER)));
    if(newPower < 0) newPower = 0;
  }else{
    newPower = min((float)-dashPower, (float)(stamina - (effortDecThreshold + Player::TIRED_BUFFER) * 2.0));
    if(newPower < 0) newPower = 0;
    newPower = -newPower;
  }

  return newPower;
}  

void Player::updatePredictedStamina(float &stamina, float &effort,
                                    float &recovery, float dashPower) const
{
  float maxStamina = Configuration::getStaminaMax(type);
  //reduce the stamina amount due to a dash
  if(dashPower > 0)
    stamina -= dashPower;
  else
    //substact the doubled amount in case of running backward
    stamina -= 2 * dashPower;
  stamina = max(stamina, (float) 0);

  //handle recovery decrement
  if(stamina <= Configuration::recoverDecThr * maxStamina  && 
     recovery > Configuration::recoverMin ){
    recovery -= Configuration::recoverDec ;
  }

  //handle effort decrement
  if(stamina <= Configuration::effortDecThr * maxStamina && 
     effort > Configuration::effortMin){
    effort -= Configuration::effortDec;
    effort = max(effort, Configuration::effortMin);
  }

  //handle effort increment
  if(stamina >= Configuration::effortIncThr * maxStamina && 
     effort < Configuration::effortInit){
    effort += Configuration::effortInc;
    effort = min(effort, Configuration::effortInit);
  }

  //increase the stamina amount (do regenerate)
  stamina += recovery * Configuration::getStaminaIncMax(type);
  stamina = min(stamina, maxStamina);
}

void Player::estimateState(Time *time){
  if (isVelGoodConfident(time)){
    position.moveBy(velocity * Configuration::getPlayerDecay(type));
  }
}

int Player::predictedCyclesToPoint(const Vector2D &point, bool goalie) const {
  // const float GOALIE_ADV   = 2.0;
  // const float ACCELERATION = 0.6;
 
  float actualSpeedMax = Configuration::getMaximumPlayerSpeed(type);
  float pointDist      = position.distanceTo(point);

  if(goalie) pointDist -= Configuration::catchableAreaL; // GOALIE_ADV;
  else       pointDist -= Configuration::getPlayerSize(type) + 
	       Configuration::ballSize + 
	       Configuration::getKickableMargin(type);

  int predCycles = (int)(pointDist / actualSpeedMax);

  //penalize a unfovarable direction with 1-2 cycles
  // only penalize unfavorable direction if he is further away
  int turnCycles = 0;
  if (predCycles <= 1) return predCycles;
  if(!isnan(bodyDir) && !isinf(bodyDir)){
    float angleToPoint = fabs(normalizeTo180Deg((point - position).getDirection() - bodyDir));
    if     (angleToPoint >= 160) turnCycles = 2;
    else if(angleToPoint >=  20) turnCycles = 1;
  }

  return predCycles + turnCycles;
}















