
/*****************************************************************************
 * WorldModel_HighLevel.cc
 * Copyright (c) 2001 Virtual Werder
 *
 * implements:
 *   class WorldModel
 *
 * requires:
 *
 * Created on: 25.03.2001 by wenns
 * Maintainer: harlekin
 * Group:
 * Co-authors: asm, mag, wollny, janw, cyco, dbernau
 *
 * Changelog:
 *
 * $Id: WorldModel_HighLevel.cc,v 1.76 2004/06/29 10:57:45 mosh Exp $
 *
 ****************************************************************************/

#include "WorldModel.hh"
#include "debug.hh"
#include "common.hh"
#include "PotentialField.hh"
#include <algorithm>

//-----------------  number of players in areas  --------------------
void WorldModel::addUpInsidePlayers(const Rectangle &rect,
				    const vector<Player> *guys,
				    float conf,
				    int *number) const {
  static vector<Player>::const_iterator i;
  for(i = guys->begin(); i!=guys->end(); ++i){
    if(i->getConfidence(&time) < conf)
      continue;
    if(rect.isInside(i->getPosition()))
      (*number)++;
  }
}

/*************************************************************************/

int WorldModel::numInSetInRectangle(const PlayerSet ps, 
				    const Rectangle &rect,
				    float conf) const {
  if(conf == -1) conf = ConfigFile::goodConfidencePlayer;

  const vector<Player> *mates, *opps, *unknowns;
  setUpPlayerSets(ps, mates, opps, unknowns);
  
  int number = 0;
  if(mates   !=NULL)  addUpInsidePlayers(rect, mates,    conf, &number);
  if(opps    !=NULL)  addUpInsidePlayers(rect, opps,     conf, &number);
  if(unknowns!=NULL)  addUpInsidePlayers(rect, unknowns, conf, &number);
  
  return number;
}

/*************************************************************************/

void WorldModel::addUpInsidePlayers(const Circle &circle,
				    const vector<Player> *guys,
				    float conf,
				    int *number) const {
  static vector<Player>::const_iterator i;
  for(i = guys->begin(); i!=guys->end(); ++i){
    if(i->getConfidence(&time) < conf)
      continue;
    if(circle.isInside(i->getPosition()))
      (*number)++;
  }
}

/*************************************************************************/

int WorldModel::numInSetInCircle(const PlayerSet ps, 
				 const Circle &circle,
				 float conf) const {
  if(conf == -1) conf = ConfigFile::goodConfidencePlayer;

  const vector<Player> *mates, *opps, *unknowns;
  setUpPlayerSets(ps, mates, opps, unknowns);
  
  int number = 0;
  if(mates   !=NULL)  addUpInsidePlayers(circle, mates,    conf, &number);
  if(opps    !=NULL)  addUpInsidePlayers(circle, opps,     conf, &number);
  if(unknowns!=NULL)  addUpInsidePlayers(circle, unknowns, conf, &number);
  
  return number;
}

/*************************************************************************/

void WorldModel::addUpInsidePlayers(const Cone &cone,
				    const vector<Player> *guys,
				    float conf,
				    int *number) const {
  static vector<Player>::const_iterator i;
  for(i = guys->begin(); i!=guys->end(); ++i){
    if(i->getConfidence(&time) < conf)
      continue;
    if(cone.isInside(i->getPosition()))
      (*number)++;
  }
}

/*************************************************************************/

int WorldModel::numInSetInCone(const PlayerSet ps,
			       const Cone &cone,
			       float conf) const {
  if(conf == -1) conf = ConfigFile::goodConfidencePlayer;

  const vector<Player> *mates, *opps, *unknowns;
  setUpPlayerSets(ps, mates, opps, unknowns);
  
  int number = 0;
  if(mates   !=NULL)  addUpInsidePlayers(cone, mates,    conf, &number);
  if(opps    !=NULL)  addUpInsidePlayers(cone, opps,     conf, &number);
  if(unknowns!=NULL)  addUpInsidePlayers(cone, unknowns, conf, &number);
  
  return number;
}


/*************************************************************************/

void WorldModel::addUpInsidePlayers(const Vector2D &start,
				    const Vector2D &end,
				    double width,
				    const vector<Player> *guys,
				    float conf,
				    int *number) const{
  
  static Vector2D posOnLine, posObj;
  Line line = Line::makeLineFromTwoPoints(start,end);

  static vector<Player>::const_iterator i;
  for(i = guys->begin(); i!=guys->end(); ++i){
    if(i->getConfidence(&time) < conf)
      continue;
    posObj = i->getPosition();
    posOnLine = line.getPointOnLineClosestTo( posObj );
    if (posOnLine.distanceTo( posObj ) < 
	   width * posOnLine.distanceTo( start )
	&& line.isInBetween( posOnLine, start, end )
	&& start.distanceTo( posObj ) < start.distanceTo( end ) )
      (*number)++;
  }
  
}
				   

/*************************************************************************/

/*TEST*/ int WorldModel::numInSetInCone(const PlayerSet ps,
					const Vector2D &start,     // 
					const Vector2D &end,       // 
					double width,       // width of cone after 1.0
					float conf) const {
  
  if (conf == -1) conf = ConfigFile::goodConfidencePlayer;
  int number = 0;

  const vector<Player> *mates, *opps, *unknowns;
  setUpPlayerSets(ps, mates, opps, unknowns);
  
  if(mates   !=NULL) addUpInsidePlayers(start, end, width, mates,    conf, &number);
  if(opps    !=NULL) addUpInsidePlayers(start, end, width, opps,     conf, &number);
  if(unknowns!=NULL) addUpInsidePlayers(start, end, width, unknowns, conf, &number);
  
  return number;
}
//-----------------  /number of players in areas  --------------------


//--------------------------- is/being attacked -----------------------
/*TEST*/ bool WorldModel::isAttacked(const Player &player) const {
  if (player == myself){
    const Player *enemy = closestInSetToPlayer(PS_Opponents, player, ConfigFile::highConfidencePlayer);
    if (enemy == NULL) return false;

    // Enemy is close, use complex method
    if (!enemy->isClose()){
      float distance = enemy->getPosition().distanceTo(ball.getPosition());
      float kickMargin = Configuration::getKickableMargin(enemy->getType()) + 
	Configuration::getPlayerSize(enemy->getType());
      float catchMargin = Configuration::catchableAreaL;
      float ballSize = Configuration::ballSize;
      float enemyAccelMax = 100 * 
	Configuration::getDashPowerRate(enemy->getType()) * 
	Configuration::getEffortMax(enemy->getType());
      float bothPlayerSizes = Configuration::getPlayerSize(myself.getType()) +
	Configuration::getPlayerSize(enemy->getType());
      
      // Case 1: attacked this turn
      float margin;
      if (enemy->isGoalKeeper()) 
	margin = catchMargin;
      else
	margin = kickMargin;
      
      if (distance < margin + ballSize) 
	return true;
      
      // Case 2: attacked next turn
      // Calculate extreme points which he can reach
      Vector2D maxPoint = enemy->predictedPosition() + Vector2D( enemyAccelMax, enemy->getBodyDir(), POLAR);
      Vector2D minPoint = enemy->predictedPosition() + Vector2D(-enemyAccelMax, enemy->getBodyDir(), POLAR);
      Line line = Line::makeLineFromTwoPoints(maxPoint, minPoint);
      
      // Determine closest point to ball
      Vector2D closestPoint = line.getPointOnLineClosestTo(ball.predictedPosition());    
      Vector2D rightPoint = maxPoint.getX() > minPoint.getX() ? maxPoint : minPoint;
      Vector2D leftPoint  = maxPoint.getX() < minPoint.getX() ? maxPoint : minPoint;
      
      // Determine which point between those extreme points is closest to the point
      static Vector2D point;
      if (line.isInBetween(closestPoint, rightPoint, leftPoint)){
	point = closestPoint;
      }else{
	if (closestPoint.isInFrontOf(rightPoint)){
	  point = rightPoint;
	}else{
	  point = leftPoint;
	}
      }
      
      // Ball too far away
      if (point.distanceTo(ball.predictedPosition()) > margin + ballSize) 
	return false;
      
      // Would collide with player
      if (point.distanceTo(myself.predictedPosition()) <= bothPlayerSizes)
	return false;
      
      return true;
    }
  }
  return (numInSetInCircle(PS_Opponents, Circle(player.getPosition(), IS_ATTACKED_RADIUS), ConfigFile::goodConfidencePlayer));
}

/*************************************************************************/

/** Returns if there is a Player right in front of the Player given
    as argument. It is looked only for opposing Players, that are in a
    Rectangle 10,10 in front of the Player wanting to dribble.
    @return true, if the Player is Frontal Attacked, false otherwise
    @todo this method is not TESTED yet, use on your own RISK
*/
bool WorldModel::isFrontalAttacked(const Player& player) const {

  //player position on the field
  const Vector2D &playerPos = player.getPosition();
  static Vector2D vec_LT, vec_RB;

  //creating the frontal_attacked_rectangle 
  if(player.getBodyDir() != UNDEFINED_FLOAT){
    vec_LT = playerPos + Vector2D(FRONTAL_ATTACKED_WIDTH/2, player.getBodyDir() - 90, POLAR);
    vec_RB = playerPos + Vector2D(FRONTAL_ATTACKED_HEIGHT, player.getBodyDir(), POLAR) + 
                         Vector2D(FRONTAL_ATTACKED_WIDTH/2, player.getBodyDir() + 90, POLAR);
  }else{
    vec_LT = Vector2D(playerPos.getX(), playerPos.getY() - FRONTAL_ATTACKED_WIDTH/2);
    vec_RB = Vector2D(playerPos.getX() + FRONTAL_ATTACKED_HEIGHT, playerPos.getY() + FRONTAL_ATTACKED_WIDTH/2);
  } 
  
  Rectangle rec(vec_LT, vec_RB);

  //true, if opponent(s) in rectangle, false otherwise
  return(numInSetInRectangle(PS_Opponents, rec, ConfigFile::goodConfidencePlayer));
}

/*************************************************************************/

/** Returns the Point to which it is best to dribble because it is
    the direction in which the next opponent is farest Away. Returning
    the same Vector as the current Position if no changes were made. 
    Dribbling will only take place up to 10 meters to the outline.
    @param player the Player who shall perform the dribble
*/
void WorldModel::getBestDribblePoint(const Player &player, Vector2D &drPt){
  // float ownX = player.getPosition().getX();
  // float ownY = player.getPosition().getY();
  /*
   * initialize dribblepoint
   */
  /*
  Vector2D upperPoint, centerPoint, lowerPoint;
  float upperDist, centerDist, lowerDist;
  upperPoint  = Vector2D(42,-15);
  centerPoint = Vector2D(40,  0);
  lowerPoint  = Vector2D(42, 15);
  Vector2D goalMid         = field.getMarkerPosition(GOAL_R);
  Vector2D leftGoalCorner  = field.getMarkerPosition(FLAG_GRT);
  Vector2D rightGoalCorner = field.getMarkerPosition(FLAG_GRB);
  // dont kick to the posts
  leftGoalCorner.setY(leftGoalCorner.getY() + 1);
  rightGoalCorner.setY(rightGoalCorner.getY() - 1);

  upperDist  = isNotIntersected_NH(leftGoalCorner, upperPoint, 2.0);
  centerDist = isNotIntersected_NH(goalMid, centerPoint, 2.0);
  lowerDist  = isNotIntersected_NH(rightGoalCorner, lowerPoint, 2.0);

  if (upperDist  != 0 && upperDist == max(max(upperDist, centerDist), lowerDist))
    drPt = upperPoint;
  else if (lowerDist  != 0 && lowerDist == max(max(upperDist, centerDist), lowerDist))
    drPt = lowerPoint;
  else
    drPt = centerPoint;
  return;
  */
  /* old */


  Vector2D dribblePoint(0,0,CARTESIAN);
  Vector2D dribblePtTemp(0,0,CARTESIAN);
  
  
  /*
   * add the negative gradients of the playfield, enemy and our goal for the playerposition to the dribblepoint
   */
  potField.playFieldBorderGradient(player.getPosition(), dribblePtTemp);
  dribblePoint+=dribblePtTemp;
  potField.theirGoalLogisticGradient(player.getPosition(), dribblePtTemp);
  dribblePoint+=dribblePtTemp;
  potField.ourGoalLogisticGradient(player.getPosition(), dribblePtTemp);
  dribblePoint+=dribblePtTemp;
#define OUTDRIBBLE_KNOWNENEMIES
#ifdef OUTDRIBBLE_KNOWNENEMIES
  static vector<Player>::iterator i;
  /*
   * add the negative gradients for each known enemy player to the dribblepoint
   */
  for(i=opponents->begin(); i < opponents->end(); ++i){
    if(!i->isGoodConfident(&time)){
      continue;
    }
    potField.theirPlayerGaussGradient(player.getPosition() - i->getPosition(), dribblePtTemp);
    dribblePoint += dribblePtTemp;
  }
#endif
#ifdef OUTDRIBBLE_UNKNOWNPLAYERS
  /*
   * add the negative gradients for each unknown player to the dribblepoint
   * (just because we dont know whether it is an enemy or it wont become less dangerous ;))
   */
  for(i=unknownPlayers.begin(); i < unknownPlayers.end(); ++i){
    if(!i->isGoodConfident(&time)){
      continue;
    }
    potField.theirPlayerGaussGradient(player.getPosition() - i->getPosition(), dribblePtTemp);
    dribblePoint += dribblePtTemp;
  }
  /*
   * at this point we got the negative gradient of all potentialobjects
   * as a dribblepoint
   * FIXME: since the gradient only contains useful info about the direction,
   * we should do something to get a reasonable distance...
   */
#endif
  /*
   * normalize...
   */
  dribblePoint/=dribblePoint.getLength();
  /*
   * one step length...
   */
  dribblePoint*=-5;
  /*
   * since all the gradients represent relative position changes, add own position
   * to get an actual dribble target...
   */
  dribblePoint+=player.getPosition();
  // FOOLISH, just testing stuff
  drPt.setVector2D(dribblePoint.getX(), dribblePoint.getY());
}

//--------------------------- /is/being attacked ----------------------


//-------------------------- misc predicates --------------------------
bool WorldModel::isBallCatchable() const {
  return ( Configuration::catchableAreaL >= myself.getPosition().distanceTo(ball.getPosition()) && 
	   Field::OWN_PENALTY_TOLERANCE_AREA.isInside( ball.getPosition() ) );
}

/*************************************************************************/
bool WorldModel::isBallKickable() const { 
  return ( Configuration::getKickableMargin(myself.getType()) > 
	 (distanceToBall() - Configuration::getPlayerSize(myself.getType()) - Configuration::ballSize) );
}

/*************************************************************************/
bool WorldModel::willBallBeKickable(int cycles, float dashPower) const {
  DEBUG5_HARLEKIN("willBallBeKickable(", cycles, ",", dashPower, ")");
  Vector2D ballPredictedPosition = ball.predictedPosition(cycles);
  Vector2D myPredictedPosition   = myself.predictedPosition(cycles, dashPower);
  
  return (distanceFromPointToPoint(ballPredictedPosition, myPredictedPosition)
          - Configuration::getPlayerSize(myself.getType()) - Configuration::ballSize)
          < Configuration::getKickableMargin(myself.getType()) ? true : false;
}

/*************************************************************************/
bool WorldModel::isGoalShot() const{
  Vector2D ballPos = ball.getPosition();
  float ballDir;
  if (!ball.getVelocity().isInvalid() && ball.getVelocity().getLengthSq() > 0){
    ballDir = ball.getVelocity().getDirection();
  }else{
    // if we dont know about the velocity expect the worst
    // but only if the ball is inside the penalty area
    // otherwise we would get it all the time
    return Field::OWN_PENALTY_TOLERANCE_AREA.isInside(ballPos); 
  }
  float maxGoalError = ConfigFile::goalShotMaxGoalError;

  //test: predicted postion of the ball in a number(currently 50) of cycles
  //is behind the own goalline
  if( ball.predictedPosition(ConfigFile::lookAheadCycles).getX() > -Field::LENGTH/2 )
    return false;

  Line shotLine = Line::makeLineFromPositionAndAngle(ballPos, ballDir);
  float cross = shotLine.getYGivenX(-Field::LENGTH/2.0f);
  if (fabs(cross) < (Field::GOAL_WIDTH/2.0f + maxGoalError))
    return true;
  return false;
}

/*************************************************************************/

/*REIMPLEMENT*/ bool WorldModel::isAGoodPointToScore(const Vector2D& point, bool oppGoal) {
  if (point.isInvalid())
    return false;

  if (!oppGoal){
    static Vector2D best;
    best = getBestPointToShootTo(point);
    if (!best.isInvalid()){
      DEBUG2_HARLEKIN("DEBUG: Is a good point to score: ", best);
      return true;
    }else
      return false;
  }else{
  //because its imho to simple

  /*
  if (point.getX() < 35) return false;
  if (fabs(point.getY()) > 22) return false;
  return true;
  */

  /*
  float kickedPower      = 2.7; // check for Direction!!!
  float ballSpeedMax   = Configuration::ballSpeedMax;
  float ballSpeedDecay = Configuration::ballDecay;
  const Player *goalie   = theirTeam.getGoalie();
  
  bool posTop = true, posBottom = true, posMid = true;

  Vector2D oppGoalTop    = field.getMarkerPosition(FLAG_GRT);
  Vector2D oppGoalBottom = field.getMarkerPosition(FLAG_GRB);
  Vector2D oppGoalMid    = field.getMarkerPosition(GOAL_R);
  oppGoalTop.setY(oppGoalTop.getY() + 1.5);
  oppGoalBottom.setY(oppGoalBottom.getY() - 1.5);

  Vector2D tmpVec;
  float distanceTop, distanceBottom, distanceMid;
  distanceTop    = oppGoalTop.distanceTo(point);
  distanceBottom = oppGoalBottom.distanceTo(point);
  distanceMid    = oppGoalMid.distanceTo(point);
  // cyclesTo
  int tmpCycle = 0;
  int gCycles = 0;
  // int ctGoalTop, ctGoalBottom, ctGoalMid = 1;
  float tmpDist = 0; // kickedPower;
  // tmpDist = distanceTop;
  while ( ( kickedPower * pow(ballSpeedDecay, tmpCycle) > 0.5 ) &&
	  (posTop||posMid||posBottom) &&
	  (tmpDist < distanceTop && tmpDist < distanceMid && tmpDist < distanceBottom) ){
    tmpCycle++;
    tmpDist += kickedPower * pow(ballSpeedDecay, tmpCycle - 1);
    if (posTop && tmpDist < distanceTop){
      tmpVec = oppGoalTop - point;
      tmpVec.setLength(tmpDist);
      tmpVec = point + tmpVec;
      gCycles = goalie->predictedCyclesToPoint(tmpVec, true);
      // DEBUG4_HARLEKIN(" -Cycle: ", tmpCycle, " gCycle: ", gCycles);
      if (gCycles <= tmpCycle)
	posTop = false;
    }
    if (posMid && tmpDist < distanceMid){
      tmpVec = oppGoalMid - point;
      tmpVec.setLength(tmpDist);
      tmpVec = point + tmpVec;
      gCycles = goalie->predictedCyclesToPoint(tmpVec, true);
      // DEBUG4_HARLEKIN(" -Cycle: ", tmpCycle, " gCycle: ", gCycles);
      if (gCycles <= tmpCycle)
	posMid = false;
    }
    if (posBottom && tmpDist < distanceBottom){
      tmpVec = oppGoalBottom - point;
      tmpVec.setLength(tmpDist);
      tmpVec = point + tmpVec;
      gCycles = goalie->predictedCyclesToPoint(tmpVec, true);
      // DEBUG4_HARLEKIN(" -Cycle: ", tmpCycle, " gCycle: ", gCycles);
      if (gCycles <= tmpCycle)
	posBottom = false;
    }
    // DEBUG8_HARLEKIN("Should shoot, Cycle: ", tmpCycle, " Dist: ", tmpDist, " Bools: ", posTop, posMid, posBottom);
  }
  return (posMid || posTop || posBottom);
  */


  // new method
  /*
    // PROBLEM: If no player is seen but is there... 
  if (myself.getPosition().getX() < 0) return false;

  Vector2D oppGoalTop    = field.getMarkerPosition(FLAG_GRT);
  Vector2D oppGoalBottom = field.getMarkerPosition(FLAG_GRB);
  Vector2D oppGoalMid    = field.getMarkerPosition(GOAL_R);
  oppGoalTop.setY(oppGoalTop.getY() + 1.5);
  oppGoalBottom.setY(oppGoalBottom.getY() - 1.5);

  Vector2D relBallPos = ball.getPosition() - myself.getPosition();
  float ballDist  = relBallPos.getLength();
  float ballAngle = angleToBallRel2Body();
  float ballSpeed = Configuration::maxpower * 
    ball.calculateKickRate(ballDist, ballAngle) * 
    Configuration::kickPowerRate;

  return (isIntersected_NH(oppGoalMid,    point, ballSpeed) ||
	  isIntersected_NH(oppGoalTop,    point, ballSpeed) ||
	  isIntersected_NH(oppGoalBottom, point, ballSpeed));
  */
  // old method

  
    const int MAX_GOAL_KICK_DIST = 22;// 18; // 16;
    const int MIN_GOALIE_DIST    = 7;
    
    static Vector2D oppGoalPos;
    oppGoalPos = field.getMarkerPosition(GOAL_R);
    if(oppGoal){
	oppGoalPos= field.getMarkerPosition(GOAL_L);
    }
    float goalDist   = oppGoalPos.distanceTo(point);
    const Player *goalie   = theirTeam.getGoalie();
    float goalieDist = MIN_GOALIE_DIST -1;
    if(goalie != NULL)
      goalieDist = distanceToPlayer(*goalie);
    
    if( goalDist <= MAX_GOAL_KICK_DIST && goalieDist < MIN_GOALIE_DIST)
      return true;
    return false;
  }
}

/*************************************************************************/

bool WorldModel::isAtGoodPointToScore(const Player& player, bool oppGoal){
  return isAGoodPointToScore(player.getPosition(), oppGoal); // WOW!
}
//-------------------------- /misc predicates -------------------------


//----------------------- closest to point predicates -----------------
bool WorldModel::isClosestToPoint(const Player& player,
				  const Vector2D& point,
				  float conf,
				  const Player *excludePlayer) const {
  // get the player (out of all players), who's closest to the given point
  const Player *closest = closestInSetToPoint(PS_Players, point, conf, excludePlayer);
  if ( closest == NULL )
    // there is no player...
    return false;
  else
    // return, if the given player is the closest player
    return (*closest == player);
}

/*************************************************************************/

bool WorldModel::isClosestTeammateToPoint(const Player &teammate,
					  const Vector2D &point,
					  float conf,
					  const Player *excludePlayer,
					  bool considerMayKickBall) const {
  const Player *closest = closestInSetToPoint(PS_Teammates, point, conf, excludePlayer, considerMayKickBall);
  if ( closest == NULL )
    return false;
  else
    return (*closest == teammate);
}

/*************************************************************************/

bool WorldModel::isSecondClosestToPoint(const Player &player,
					const Vector2D &point,
					float conf,
					const Player *excludePlayer) const {
  const Player *secondClosest = 
    secondClosestInSetToPoint(PS_Players, point, conf, excludePlayer);
  if ( secondClosest == NULL )
    return false;
  else
    return (*secondClosest == player);
}

/*************************************************************************/

bool WorldModel::isSecondClosestTeammateToPoint(const Player &teammate,
						const Vector2D &point,
						float conf,
						const Player *excludePlayer) const {
  const Player *secondClosest = 
    secondClosestInSetToPoint(PS_Teammates, point, conf, excludePlayer);
  if ( secondClosest == NULL )
    return false;
  else
    return (*secondClosest == teammate);

}
//----------------------- /closest to point predicates ----------------


//-----------------------  closest to point ---------------------------
void WorldModel::setUpPlayerSets(const PlayerSet ps,
				 const vector<Player> *&mates,
				 const vector<Player> *&opps,
				 const vector<Player> *&unknowns) const {
  //dependent on the set, initialiaze the vector pointer either with NULL
  //or with some other address
  switch(ps){
  case PS_Players:
    mates = teammates, opps = opponents, unknowns = &unknownPlayers;
    break;
  case PS_Teammates:
    mates = teammates; opps = NULL; unknowns = NULL;
    break;
  case PS_Opponents:
    mates = NULL, opps = opponents, unknowns = NULL;
    break;
  case PS_Unknowns:
    mates = NULL, opps = NULL, unknowns = &unknownPlayers;
    break;
  default:
    mates = NULL, opps = NULL, unknowns = NULL;
  }
}

/*************************************************************************/

void WorldModel::searchClosest(const Vector2D &point,
			       const vector<Player> *guys,
			       float conf, float *minDist,
			       const Player *&closest,
			       const Player *excludePlayer,
			       bool considerMayKickBall) const {
  float dist;
  // quadrate, so we dont need to use the sqrt-function too often
  (*minDist) = *minDist * *minDist;
  static vector<Player>::const_iterator i;
  for(i = guys->begin(); i != guys->end(); ++i){
    if(i->getConfidence(&time) < conf)
      continue;
    if (considerMayKickBall && !mayKickBall( (const Player*)(&(*i)) ) )
      continue;
    dist = i->getPositionPt()->distanceToSq(point);
    if(dist <= *minDist) {
      if ( excludePlayer == NULL ){  
	*minDist = dist, closest = (const Player*)&(*i);
      }else{
	// if i is at the same position as the excluded Player
	// then i is the excluded Player
	// Small values, dont use optimized function distanceToSq()
	if ( !(i->getPositionPt()->distanceToSq(excludePlayer->getPosition()) < 0.0001 /* 0.01 */) ){
	  *minDist = dist, closest = (const Player*)&(*i);
	}
      }
    }
  }
  (*minDist) = sqrt(*minDist);
}

/*************************************************************************/

bool WorldModel::closestTMInFrontOfPlayer(const Player& player){
  // don't look for myself
  if(myself == player){
    return false;
  }
  float myX = myself.getPositionPt()->getX();
  float otherX = player.getPositionPt()->getX();
  // we are not in front of the other Player so return false
  if(myX < otherX){
    return false;
  }
  static vector<Player>::const_iterator i;
  float myDistance = myself.getPositionPt()->distanceToSq(player.getPosition());
  for(i = myTeam.getPlayers().begin(); i < myTeam.getPlayers().end();++i){
    if(!i->isGoodConfident(&time)) continue;
    // if it is the desired Player skip him
    if(i->getUniformNumber() != player.getUniformNumber()){
      // the other Player is also in front of the Player we are looking at
      if(i->getPositionPt()->getX() > otherX){
	float otherDistance = i->getPositionPt()->distanceToSq(player.getPosition());
	// the other Player is nearer to the desired Player, so let him do the work :-)
	// we have two choices for this test:
	// 1. is smaller than, so both Players are closest in front of that Player
	// 2. is smaller than or equal, so no Player is closest in front of that Player
	// we use the first choice here
	if(otherDistance < myDistance){
	  return false;
	}
      }
    }
  }
  return true;
}


/*************************************************************************/

const Player* WorldModel::closestInSetToPoint(const PlayerSet ps,
					      const Vector2D &point,
					      float conf,
					      const Player *excludePlayer,
					      bool considerMayKickBall) const {
  if(conf == -1) conf = ConfigFile::goodConfidencePlayer;
  
  static const vector<Player> *mates, *opps, *unknowns;
  setUpPlayerSets(ps, mates, opps, unknowns);
  
  float minDist = BIG_FLOAT;
  const Player *closest = NULL;
  if(mates!=NULL)    searchClosest(point, mates,    conf, &minDist, closest, excludePlayer, considerMayKickBall);
  if(opps!=NULL)     searchClosest(point, opps,     conf, &minDist, closest, excludePlayer, considerMayKickBall);
  if(unknowns!=NULL) searchClosest(point, unknowns, conf, &minDist, closest, excludePlayer, considerMayKickBall);
  
  return closest;
}

/*************************************************************************/

void WorldModel::searchSecondClosest(const Vector2D &point,
				     const vector<Player> *guys,
				     float conf, float *minDist,
				     float *sndMinDist,
				     const Player *&closest,
				     const Player *&sndClosest,
				     const Player *excludePlayer) const {
  float dist;
  static vector<Player>::const_iterator i;
  *minDist    = *minDist    * *minDist;
  *sndMinDist = *sndMinDist * *sndMinDist;
  for(i = guys->begin(); i!=guys->end(); ++i){
    if(i->getConfidence(&time) < conf)
      continue;
    
    // this is the squared distance. We save one sqrt each cycle
    dist = i->getPositionPt()->distanceToSq(point);
    
    // case 1: i is closer to point than closest
    if ( dist < *minDist ) {
      if (excludePlayer == NULL || 
	  !(i->getPositionPt()->distanceToSq(excludePlayer->getPosition()) < 0.0001) ) {
	// i is the new closest player, closest gets sndClosest
	sndClosest = closest;
	closest    = (const Player *)&(*i);
	*sndMinDist = *minDist;
	*minDist    = dist;
      } 	  
    }    
    // case 2: i is closer to point than sndClosest, but not closer than closest
    else if ( dist < *sndMinDist ) {
      if (excludePlayer == NULL || 
	  !(i->getPositionPt()->distanceToSq(excludePlayer->getPosition()) < 0.0001) ) {
	sndClosest = (const Player *)&(*i);
	*sndMinDist = dist;
      }
    }
    
    // case 3: i is far away from point -> nothing to do
  }
  // reset values
  *minDist    = sqrt(*minDist   );
  *sndMinDist = sqrt(*sndMinDist);
}

/*************************************************************************/

const Player* WorldModel::secondClosestInSetToPoint(const PlayerSet ps,
						    const Vector2D &point,
						    float conf,
						    const Player *excludePlayer) const {
  if(conf == -1) conf = ConfigFile::goodConfidencePlayer;
  
  const vector<Player> *mates, *opps, *unknowns;
  setUpPlayerSets(ps, mates, opps, unknowns);
  
  float minDist = BIG_FLOAT, sndMinDist = BIG_FLOAT;
  const Player *closest = NULL, *sndClosest = NULL;
  
  if(mates!=NULL)
    searchSecondClosest(point, mates, conf, &minDist, &sndMinDist,
			closest, sndClosest, excludePlayer);
  if(opps!=NULL)
    searchSecondClosest(point, opps, conf, &minDist, &sndMinDist,
			closest, sndClosest, excludePlayer);
  if(unknowns!=NULL)
    searchSecondClosest(point, unknowns, conf, &minDist, &sndMinDist,
			closest, sndClosest, excludePlayer);
  
  return sndClosest;
}
//----------------------- /closest to point ---------------------------


//-----------------------  closest to Outline ---------------------------
void WorldModel::searchClosestToGoalOutLine(const bool ownGoalOutLine,
					    const vector<Player> *guys,
					    float conf, float *minDist,
					    const Player *&closest) const {
  float dist;
  static vector<Player>::const_iterator i;
  for (i = guys->begin(); i!=guys->end(); ++i) {
    if (i->getConfidence(&time) < conf)
      continue;
    if ( ownGoalOutLine )
      dist = distanceToOwnGoalOutLine( *i );
    else 
      dist = distanceToOppGoalOutLine( *i );
    if(dist < *minDist) 
      *minDist = dist, closest = (const Player*)&(*i);
  }
}


/*************************************************************************/

const Player* WorldModel::closestInSetToOwnGoalOutline(PlayerSet ps,
						       float conf) const {
  if(conf == -1) conf = ConfigFile::goodConfidencePlayer;
  
  static const vector<Player> *mates, *opps, *unknowns;
  setUpPlayerSets(ps, mates, opps, unknowns);
  
  float minDist = BIG_FLOAT;
  const Player *closest = NULL;
  if(mates   !=NULL)  searchClosestToGoalOutLine(true, mates,    conf, &minDist, closest);
  if(opps    !=NULL)  searchClosestToGoalOutLine(true, opps,     conf, &minDist, closest);
  if(unknowns!=NULL)  searchClosestToGoalOutLine(true, unknowns, conf, &minDist, closest);
  
  return closest;
}
  
/*************************************************************************/

const Player* WorldModel::closestInSetToOppGoalOutline(PlayerSet ps,
						       float conf) const {
  if(conf == -1) conf = ConfigFile::goodConfidencePlayer;
  
  static const vector<Player> *mates, *opps, *unknowns;
  setUpPlayerSets(ps, mates, opps, unknowns);
  
  float minDist = BIG_FLOAT;
  const Player *closest = NULL;
  if(mates   !=NULL)  searchClosestToGoalOutLine(false, mates,    conf, &minDist, closest);
  if(opps    !=NULL)  searchClosestToGoalOutLine(false, opps,     conf, &minDist, closest);
  if(unknowns!=NULL)  searchClosestToGoalOutLine(false, unknowns, conf, &minDist, closest);
  
  return closest;
}
//-----------------------  /closest to Outline ---------------------------


//-------- Offside -----------------------------------------------------
float WorldModel::getOppOffsideX(float conf) const {
  if(conf == -1) conf = ConfigFile::goodConfidencePlayer;

  float highestX = 0.0;
  float sndHighestX = 0.0;
  
  // find the X-Coordinate of the two opponent players nearest to the oppGoalOutLine
  const vector<Player> *mates = teammates;
  vector<Player>::const_iterator i;
  for (i = mates->begin(); i != mates->end(); ++i) {
    // if (   i->getConfidence(*(&time))   <   conf   
	   if (   i->getConfidence(&time)   <   conf 
	   || i->getUniformNumber() ==  myself.getUniformNumber())
      continue;

    if (i->getPosition().getX() < highestX) {
      sndHighestX = highestX;
      highestX = i->getPosition().getX();
    } else {
      if (i->getPosition().getX() < sndHighestX) {
	sndHighestX = i->getPosition().getX();
      }
    }
    
  }
  // is the ball nearer to the oppGoal than the second opponent player?
  if (ball.getPosition().getX() < sndHighestX &&
      ball.getPosition().getX() != UNDEFINED_FLOAT)
    // the ballposition is the offsideLine
    sndHighestX = ball.getPosition().getX();

  if (sndHighestX < 0.0)
    return sndHighestX;
  else
    return 0.0f;
    // no Ball and no opponents seen in the opponent side of the field -> no offsideline
}
float WorldModel::getOffsideX(float conf) const {
  if(conf == -1) conf = ConfigFile::goodConfidencePlayer;

  float highestX = 0.0;
  float sndHighestX = 0.0;
  
  // find the X-Coordinate of the two opponent players nearest to the oppGoalOutLine
  const vector<Player> *opps = opponents;
  static vector<Player>::const_iterator i;
  for (i = opps->begin(); i != opps->end(); ++i) {
    if (i->getConfidence(&time) < conf)
      continue;
    if (i->getPosition().getX() > highestX) {
      sndHighestX = highestX;
      highestX = i->getPositionPt()->getX();
    } else {
      if (i->getPositionPt()->getX() > sndHighestX) {
	sndHighestX = i->getPositionPt()->getX();
      }
    }
  }
  // is the ball nearer to the oppGoal than the second opponent player?
  if (ball.getPositionPt()->getX() > sndHighestX &&
      ball.getPositionPt()->getX() != UNDEFINED_FLOAT)
    // the ballposition is the offsideLine
    sndHighestX = ball.getPositionPt()->getX();

  if (sndHighestX > coach_OffsideX)
    return sndHighestX;
  else
    return coach_OffsideX;
    // no Ball and no opponents seen in the opponent side of the field -> no offsideline
}
//-------- /Offside ----------------------------------------------------



/*************************************************************************/

/** Returns the Player specified by his unum or a NULL Pointer if the Player was not found
    @param unum the Uniform Number after which is searched
    @return a Player Object of the Player specified by the overgiven Unum
*/
const Player* WorldModel::getPlayerByUnum(int unum){
  // Remove if?
  if ((*teammates)[unum-1].getUniformNumber() == unum)
    return &((*teammates)[unum-1]);
  return NULL;
}

/** Returns the Player specified by his unum or a NULL Pointer if the Player was not found
    @param unum the Uniform Number after which is searched
    @return a Player Object of the Player specified by the overgiven Unum
*/
const Player* WorldModel::getOtherTeamPlayerByUnum(int unum){
  // Remove if?
  if ((*opponents)[unum-1].getUniformNumber() == unum)
    return &((*opponents)[unum-1]);
  return NULL;
}

/*************************************************************************/

float WorldModel::calcKickSpeedToTravel( const float dist,
					 const float endSpeed ) const{
  // if endspeed is zero we have an infinite series and return the first term
  // that corresponds to the distance that has to be travelled.
  if( endSpeed < 0.0001  )
    return Geometry::getFirstInfGeomSeries( dist, Configuration::ballDecay );

  // use geometric series to calculate number of steps and with that the
  // velocity to give to the ball, we start at endspeed and therefore use
  // the inverse of the ball decay (r).
  // s = a + a*r + .. a*r^n since we calculated from endspeed (a) to
  // firstspeed, firstspeed equals a*r^n = endspeed*r^nr_steps
  float cycles = Geometry::getLengthGeomSeries( endSpeed,
                                                1.0/Configuration::ballDecay,
                                                dist );
  return calcInitSpeedFromEndSpeed( endSpeed, (int)rintf(cycles) ) ;
}

/*************************************************************************/
float WorldModel::calcKickPowerForSpeed( const float speed ) const{
  float ballDist  = distanceToBall();
  float ballAngle = angleToBallRel2Body();

  return (speed / ball.calculateKickRate(ballDist, ballAngle) /
                    Configuration::kickPowerRate);
}

/*************************************************************************/

float WorldModel::calcInitSpeedFromEndSpeed( const float endSpeed,
					     const float cycles ) const {
  // geometric series: s = a + a*r^1 + .. + a*r^n, now given endspeed = a*r^n ->
  // endspeed = firstspeed * ratio ^ length ->
  // firstpeed = endspeed * ( 1 / ratio ) ^ length
  return endSpeed * pow( 1 / Configuration::ballDecay, cycles );
}

/*************************************************************************/
/**
 * This method returns the speed that has to be given to the ball when
 * it should travel a given distance in a given numbers of cycles.
 * @param distance distance to travel
 * @param cycles nr of cycles after which the ball should have travel the distance
 * @return initial speed given to the ball to travel the given distance
 * in the given number of cycles.
 */
/*************************************************************************/
float WorldModel::calcInitSpeedForDistInCycles( const float distance,
					     const int cycles ) const {
  // calculation not correct for this values -> return 0
  // TODO: Why 0? Calling method wont know if 0 is a correct value or not!
  if(distance <= 0 || cycles <= 0)
    return 0;
  //calculation for ballDecaySum
  float ballDecaySum = 0;
  for(int i=0; i < cycles; ++i){
      ballDecaySum += pow(Configuration::ballDecay, i);
  }  
  return distance / ballDecaySum;
}

/*************************************************************************/

float WorldModel::calcActualTurnAngle( const float angle,
				       const float speed ) const {
  return angle / (1.0 + Configuration::getInertiaMoment(myself.getType()) * speed );
}


/*************************************************************************/

float WorldModel::calcAngleForTurn( const float desiredAngle,
				    const float speed ) const {
  return desiredAngle * (1.0 + Configuration::getInertiaMoment(myself.getType()) * speed );
}



/*************************************************************************/
/*CORDOKIN*/
/*TEST*/ void WorldModel::searchFastest(const vector<Player> *guys,
					const Vector2D &ballPos,
					const Player *&fastest, int &minCycles,
					float conf,
					bool considerMayKickBall) const {
  int cyclesToBall;
  static vector<Player>::const_iterator i;
  for(i = guys->begin(); i != guys->end(); ++i){
    if(i->getConfidence(&time) < conf)
      continue;
    // do not consider goalies if the ball is out of their reach
    // But consider both penalty areas because this can be needed 
    // during a penalty shootout
    if (i->isGoalKeeper())
      if (!Field::OPP_PENALTY_AREA.isInside(ballPos) && 
	  !Field::OWN_PENALTY_TOLERANCE_AREA.isInside(ballPos)) continue;

    if (considerMayKickBall && !mayKickBall( (const Player*)(&(*i)) ) )
      continue;
    cyclesToBall = i->predictedCyclesToPoint(ballPos, i->isGoalKeeper());
    if(cyclesToBall < minCycles)
      minCycles = cyclesToBall, fastest = (const Player *)&(*i);
  }
}

/*************************************************************************/

/*TEST*/ const Player* WorldModel::fastestInSetToBall(const PlayerSet ps,
						      float conf,
						      bool considerMayKickBall) const {
  if(conf == -1) conf = ConfigFile::goodConfidencePlayer;
  
  static const vector<Player> *mates, *opps, *unknowns;
  setUpPlayerSets(ps, mates, opps, unknowns);
  
  static const int MAX_LOOKAHEAD = 50;
  static Vector2D ballPos;
  int minCycles = MAX_LOOKAHEAD;
  const Player *fastest = NULL;

  for(int cycles = 0; cycles < minCycles && 
	              cycles < MAX_LOOKAHEAD; ++cycles){
    minCycles = BIG_INT;
    ballPos = ball.predictedPosition(cycles);
    if(mates!=NULL)
      searchFastest(mates,    ballPos, fastest, minCycles, conf, considerMayKickBall);
    if(opps!=NULL)
      searchFastest(opps,     ballPos, fastest, minCycles, conf, considerMayKickBall);
    if(unknowns!=NULL)
      searchFastest(unknowns, ballPos, fastest, minCycles, conf, considerMayKickBall);
  }

  return fastest;
}

/*TEST*/ const Player* WorldModel::fastestInSetToPoint(const PlayerSet ps,
						       Vector2D pos,
						       int minCycles, 
						       float conf,
						       bool considerMayKickBall) const {
  if(conf == -1) conf = ConfigFile::goodConfidencePlayer;
  
  static const vector<Player> *mates, *opps, *unknowns;
  setUpPlayerSets(ps, mates, opps, unknowns);
    
  const Player *fastest = NULL;

  //  for(int cycles = 0; cycles < minCycles; cycles++){
  //minCycles = BIG_INT;
    if(mates!=NULL)
      searchFastest(mates, pos, fastest, minCycles, conf, considerMayKickBall);
    if(opps!=NULL)
      searchFastest(opps, pos, fastest, minCycles, conf, considerMayKickBall);
    if(unknowns!=NULL)
      searchFastest(unknowns, pos, fastest, minCycles, conf, considerMayKickBall);
    //}

  return fastest;
}

void WorldModel::resetPrecalculatedValues(){
  iP_calculated               = false;
  sP_calculated               = false;
  pre_isClosestTeammateToBall = false;
}

/*************************************************************************/

bool WorldModel::interceptionPoint(const Player& p, Vector2D &intPoint,
                                   int &noCycles, bool goalie) {
  if (!iP_calculated){
    iP_calculated = true;
    iP_intable = false;
    for(int i = 0; i < Player::MAX_LOOKAHEAD; ++i){
      iP_intPoint = ball.predictedPosition( i );
      
      // if the ball position is farther away then we can run 
      // (use some huge tolerance) take the next position
      /*
      if (p.getPositionPt()->distanceToSq(iP_intPoint) > ((float)(i * i) * 2.00f))
	continue;
      */
      iP_noCycles = p.predictedCyclesToPoint(iP_intPoint, goalie);
      // i >= noCycles arrive at the same time as the ball
      // i >  noCycles arrive before the ball does
      if(i > iP_noCycles){
	iP_intable = true;
	break;
      }
    }
  }
  noCycles = iP_noCycles;
  intPoint = iP_intPoint;
  return iP_intable;
}

/*************************************************************************/

bool WorldModel::safetyPoint(const Player& p, Vector2D &safetyPoint,
                             int &noCycles, bool goalie) {
  if (!sP_calculated){
    sP_calculated = true;
    //calculate interception point if not calculated yet
    if (!iP_calculated){
      Vector2D temp;
      int tempCycles;
      interceptionPoint(p, temp, tempCycles, goalie);
    }
    //there can only be a safety point, when there is an
    //interception point 
    if (iP_intable){
      int lastCheckCycles = BIG_INT;
      Vector2D newPoint;
      int newCycles;
      for(int i = 0; i < Player::MAX_LOOKAHEAD; i++){
        //safetyPoint
	newPoint = ball.predictedPosition( i );
	newCycles = p.predictedCyclesToPoint(newPoint, goalie);
        //where the agent needs the lowest cycles to point
        if (lastCheckCycles < newCycles){
	  noCycles = sP_noCycles;
          break;
	}
	sP_noCycles = newCycles;
	safetyPoint = newPoint;
	lastCheckCycles = sP_noCycles;
      }
      sP_intPoint = safetyPoint;
      sP_intable  = true;
      return true;
    }
    sP_intable  = false;
    return false;
  }
  noCycles    = sP_noCycles;
  safetyPoint = sP_intPoint;
  return sP_intable;
}
/*************************************************************************/

bool WorldModel::isMyHomePosClosestToPos(const Vector2D &pos) const {
  int unum      = UNDEFINED_INT;
  float minDist = BIG_FLOAT;
  float currentDist;
  const Formation *f = Formations::getCurrentFormation();
  for(int i = 2; i <= 11; ++i){   //do not consider the goalie
    currentDist = pos.distanceToSq(f->getHomePos(i));
    if( currentDist < minDist ){
      unum = i;
      minDist = currentDist;
    }
  }

  if(unum == myself.getUniformNumber())
    return true;

  return false;
}

/*************************************************************************/

// TODO: Rewrite method. It is not accurate anymore as the positioning has changed
Vector2D WorldModel::calcSBSPPosition(int unum){
  Playmode p = /*situation.*/getPlaymode();
  Vector2D pos = Formations::getCurrentFormation()->getHomePos(unum);
  Vector2D ballPos = ball.getPosition();
  Vector2D relBallPos = ballPos - pos;

  //ball attraction factors
  float factorX = Formations::getCurrentFormation()->getAttractionX(unum);
  float factorY = Formations::getCurrentFormation()->getAttractionY(unum);

  //dependence on the ball position
  pos.setX(pos.getX() + relBallPos.getX()*factorX);
  pos.setY(pos.getY() + relBallPos.getY()*factorY);

  //avoid the boundaries of the field
  if (fabs(pos.getY()) > 30.0) pos.setY(pos.getY() - 0.5*factorY*relBallPos.getY());

  //avoid offside positions?
  //Position = PositionToOnsidePosition(Position);

  //make sure the defence goes behind the ball
  if (Formations::getCurrentFormation()->getBehindBall(unum) && 
      ballPos.getX() < pos.getX()) pos.setX(ballPos.getX());

  //make sure no one goes closer to the enemies goal as the ball if the enemy
  if(p == PM_OurOffside) pos.setX(ballPos.getX());

  //monitor an maximum and a minimum of the x-coord of the strategic position
  pos.setX(min((double)pos.getX(), Formations::getCurrentFormation()->getMaxX(unum)));
  pos.setX(max((double)pos.getX(), Formations::getCurrentFormation()->getMinX(unum)));

  return pos;
}

/*************************************************************************/
//TEST
float WorldModel::getDashPowerForPositioning(){
    int type = myself.getType();
    float dashPower, dashPowerStamina, dashPowerBallDist, dashPowerPosDist;
    float reserve = (Configuration::getStaminaMax(type) * Configuration::effortDecThr) + Player::TIRED_BUFFER;

    //three calculations for stamina sparing dashPower

    //calculation for the current stamina situation
    float staminaNextCycle = min( Configuration::getStaminaMax(type), (Configuration::getStaminaIncMax(type) +
				  myself.getStamina()) );
    dashPowerStamina = ( (staminaNextCycle - reserve) / (Configuration::getStaminaMax(type) - reserve) );
    
    //calculation for the current distance to the ball
    float ballDist = distanceToBall();
    if( ballDist < 12.5 ){
      dashPowerBallDist = 1.0;
    }else if( ballDist > 50.0 ){
      dashPowerBallDist = 0.0;
    }else{
      dashPowerBallDist = 1.0 - ( (ballDist - 12.5) / 37.5 );
    }
    
    //calculation for the current distance to the positing point
    Vector2D positioningPoint = calcSBSPPosition( myself.getUnum() );
    dashPowerPosDist = min( distanceToPoint( positioningPoint ) , (float)30.0 ) /
		       (float)30.0;
    DEBUG4(" dashPowerPosDist; ", dashPowerPosDist,
	   " bei : ", distanceToPoint( positioningPoint ));

    //final calculation to ut the three ascepts together
    dashPower = Configuration::getStaminaIncMax(type) + 
	        min( dashPowerStamina, max( dashPowerBallDist, dashPowerPosDist ) ) * 
	        ( Configuration::maxpower - Configuration::getStaminaIncMax(type) );
    return dashPower;
}


//----------------------------------------------------------------------
//pass related routines
//----------------------------------------------------------------------
/*************************************************************************/
bool WorldModel::isNotIntersected_PIOS(Vector2D &ballVelocity,
				       const Vector2D &start,
				       const Vector2D &goal,
				       const Vector2D &ourPos){
  //Line line = Line::makeLineFromTwoPoints(start,goal);
  Vector2D lineVector = goal - start;
  float length = lineVector.getLength();

  static vector<Player>::iterator i;
  float tolerance = 0.1;
  
  Vector2D travelledDistance = ballVelocity;
  
  int count = 0;

  while(length > travelledDistance.getLength()){ 
   
      for (i = theirTeam.getPlayers().begin(); i < theirTeam.getPlayers().end(); ++i) {
	if (!i->isGoodConfident(&time)) continue;
	float distToPoint = (start+travelledDistance).distanceTo( (i->getPosition()+i->getVelocity()) );
	int type = i->getType();
	if(i->isGoalKeeper()){
	  if(distToPoint<((count) * Configuration::getPlayerSpeedMax(type) + 
			  Configuration::catchableAreaL + (tolerance)))
	    return false;
	}
	else
	  if(distToPoint<((count) * Configuration::getPlayerSpeedMax(type) + 
			  Configuration::getKickableMargin(type) + tolerance))     
	    return false;	    
      }
      float distToPoint = (start+travelledDistance).distanceTo(ourPos);
      if(distToPoint < ((count-(3 + int(travelledDistance.getLength()/5) ))*Configuration::playerSpeedMax+Configuration::kickableMargin + tolerance))
	return true;
      
      
      count++;
      ballVelocity = ballVelocity * Configuration::ballDecay;
      travelledDistance += ballVelocity;     
    }
  return true;
}

/*************************************************************************/
float WorldModel::intersectionRisk(const Vector2D &from, 
				   const Vector2D &to, vector<Player*> &ignore) {
  if (from == to) return 0.0;
  float risk = 0.0;
  Line passLine = Line::makeLineFromTwoPoints(from, to);
  float passDist = distanceFromPointToPoint(from, to);
  static vector<Player*> allPlayers;
  allPlayers.clear();
  static vector<Player>::iterator i;
  for (i = opponents->begin(); i < opponents->end(); ++i) { 
    if (*(find(ignore.begin(), ignore.end(), &(*i))) == *ignore.end())
      allPlayers.push_back(&(*i));
  }
  for (i = teammates->begin(); i < teammates->end(); ++i) {
    if (*(find(ignore.begin(), ignore.end(), &(*i))) == *ignore.end())
      allPlayers.push_back(&(*i));
  }
  for (i = unknownPlayers.begin(); i < unknownPlayers.end(); ++i) {
    if (*(find(ignore.begin(), ignore.end(), &(*i))) == *ignore.end())
      allPlayers.push_back(&(*i));
  }
  vector<Player*>::iterator p;
  for (p = allPlayers.begin(); p < allPlayers.end(); ++p) {
    Vector2D pos = (*p)->getPosition();
    Vector2D icP = passLine.getPointOnLineClosestTo(pos);
    float fromDist = distanceFromPointToPoint(from, icP);
    float toDist = distanceFromPointToPoint(to, icP);
    if ((fromDist > passDist) || (toDist > passDist)) {
      if (fromDist > toDist) icP = to;
      else icP = from;
    }
    fromDist = distanceFromPointToPoint(from, icP);
    float intDist = distanceFromPointToPoint(pos, icP);
    if (intDist < 0.01) {
      for (p = allPlayers.begin(); p < allPlayers.end(); ++p) {
	delete *p;
      }
      return BIG_FLOAT;
    }
    cout << "runtime: TEST" << endl;
    float quot = fromDist / intDist;
    risk += (quot*0.1)*(quot*0.1)*(quot*0.1)*(quot*0.1);
  }
  for (p = allPlayers.begin(); p < allPlayers.end(); ++p) {
    delete *p;
  }
  return risk;
}

//*********************************************************************************
bool WorldModel::isIntersected(Player &p1, 
			       Player &p2, vector<Player*>* ignore){
  bool deleteIt = false;
  static vector<Player*>::iterator i;
  if (ignore == NULL) {
    ignore = new vector<Player*>();
    deleteIt = true;
  }
  ignore->push_back(&p1);
  ignore->push_back(&p2);
  Vector2D pos1 = p1.getPosition();
  Vector2D pos2 = p2.getPosition();
  float risk = intersectionRisk(pos1, pos2, *ignore);
  if (deleteIt) {
    for (i = ignore->begin(); i < ignore->end(); ++i) {
      delete *i;
    }
    delete ignore;
  }
  return (risk >= ConfigFile::intersectionMin);
}

//*********************************************************************************
bool WorldModel::isIntersected(Player &p, vector<Player*>* ignore){
  bool deleteIt = false;
  vector<Player*>::iterator i;
  if (ignore == NULL) {
    ignore = new vector<Player*>();
    deleteIt = true;
  }
  ignore->push_back(&p);
  Vector2D pos1 = ball.getPosition();
  Vector2D pos2 = p.getPosition();
  float risk = intersectionRisk(pos1, pos2, *ignore);
  if (deleteIt) { 
    for (i = ignore->begin(); i < ignore->end(); ++i) {
      delete *i;
    }
    delete ignore;
  }
  return (risk >= ConfigFile::intersectionMin);
}

//*********************************************************************************
bool WorldModel::isIntersected(const Vector2D &pp){
  Vector2D ball_pos = ball.getPosition();
  //return isIntersected(pp, ball_pos);
  return isIntersected(pp, ball_pos);
}



/** Tests if a a Shoot is intersected
    @param p1 the Point to which is looked
    @param p2 the Point from which is looked
    @param ballSpeed the Speed of the Ball
    @return -1 if intersected
    @ else returns closest Distance of Enemys To Ball
*/

double WorldModel::isIntersected_NH_Risk(const Vector2D &p1 /*to*/, const Vector2D &p2 /*from*/, float ballSpeed){
  
  Vector2D ballWayVector = p1 - p2;
  float length = ballWayVector.getLength();
  
  static vector<Player>::iterator i;
  //float tolerance = 0.1;
  
  Vector2D ballVel;
  ballVel.setVector2D(ballSpeed, ballWayVector.getDirection(), POLAR);
  Vector2D travelledDistance = ballVel;

 for (i = theirTeam.getPlayers().begin(); i < theirTeam.getPlayers().end(); ++i) {
      if (!i->isGoodConfident(&time)) continue;
    
      Vector2D pos= i->getPosition() + i->getVelocity();
      float distToPoint = (p2+travelledDistance).distanceTo(pos);
      int type = i->getType();

      float enemysWay =(0 * Configuration::getPlayerSpeedMax(type) + 
			Configuration::getKickableMargin(type));
      
      if(i->isGoalKeeper())
	enemysWay =(0 * Configuration::getPlayerSpeedMax(type) 
		    + Configuration::catchableAreaL);
      
      if(distToPoint < enemysWay){
	return -1.0;
      }
 }
  
  
  travelledDistance += ballVel * Configuration::ballDecay;
  int cycles = 0;
  float closestEnemyDist=99;
  
  while(length > travelledDistance.getLength()){ 
    
      if(ballVel.getLength() < 0.01 || cycles > 20){
	return -1.0;
      }
      
    for (i = theirTeam.getPlayers().begin(); i < theirTeam.getPlayers().end(); ++i) {
      if (!i->isGoodConfident(&time)) continue;
    
      Vector2D pos= i->getPosition() + i->getVelocity();
      float distToPoint = (p2+travelledDistance).distanceTo(pos);
      int type = i->getType();

      float enemysWay =((cycles-int(cycles/10)) * Configuration::getPlayerSpeedMax(type) + 
			Configuration::getKickableMargin(type));
      
      if(i->isGoalKeeper())
	enemysWay =((cycles-int(cycles/10)) * Configuration::getPlayerSpeedMax(type) 
		    + Configuration::catchableAreaL);
      
      if(distToPoint < enemysWay){
	return -1.0;
      }

      float missingDistance = distToPoint - enemysWay;
      if( missingDistance < closestEnemyDist)
	closestEnemyDist = missingDistance;
      

    }
    cycles++;
    ballVel = ballVel * Configuration::ballDecay;
    travelledDistance += ballVel; 
  }
  
   
  if(closestEnemyDist < 0)
    return -1.0;
  
  
  return closestEnemyDist;

}

// p2 = myPos/ballPos etc (from), p1 = target (to)
/** Tests whether a a Point is intersected
    @param p1 the Point to which is looked
    @param p2 the Point from which is looked
    @param ballSpeed the Speed of the Ball
    @return true if intersected, false otherwise
*/


bool WorldModel::isIntersected_NH(const Vector2D &p1 /*to*/, const Vector2D &p2 /*from*/, float ballSpeed){
  static vector<Player>::iterator i;
  static Line line = Line(0,0,0);
  line.updateLineFromTwoPoints(p1,p2);
  for (i  = theirTeam.getPlayers().begin(); 
       i != theirTeam.getPlayers().end(); ++i){

    if (!i->isGoodConfident(&time)) continue;
    float bSpeed = ballSpeed;
    float dist   = ballSpeed;

    // calculate his best interception point
    Vector2D iP = line.getPointOnLineClosestTo(i->getPosition());
    if (!line.isInBetween(iP, p1, p2)) continue;
    // how long does he need to get there?
    int predictedCycles = i->predictedCyclesToPoint(iP, i->isGoalKeeper());
    // how long does the ball need?
    for (int j = 1; j <=  predictedCycles; ++j)
      dist += (bSpeed *= Configuration::ballDecay);
    // who is faster?
    if (dist <= distanceFromPointToPoint(p2, iP))
      return true;
  }
  return false;
}
/** tests wether a pass to a specific point is intercepted or not
 *
 * return: smallest distance of an enemy to the line (from person to point)
 *         0 = isIntercepted
 */
float WorldModel::isNotIntersected_NH(const Vector2D &p1, const Vector2D &p2, float ballSpeed){
  float reachableArea;
  float smallestDistance = BIG_FLOAT;
  static vector<Player>::iterator i;
  Line line = Line::makeLineFromTwoPoints(p1,p2);
  for (i = theirTeam.getPlayers().begin(); i < theirTeam.getPlayers().end(); ++i){
    if (!i->isGoodConfident(&time)) continue;
    float bSpeed = ballSpeed;
    float dist = ballSpeed;
    Vector2D iP = line.getPointOnLineClosestTo(i->getPosition());
    if (!line.isInBetween(iP, p1, p2)) continue;
    float iP2i = distanceFromPointToPoint(iP, i->getPosition());
    if(i->getUniformNumber() == 1){
      reachableArea = Configuration::catchableAreaL;
    }else{
      reachableArea = Configuration::getPlayerSize(myself.getType()) + 
	              Configuration::ballSize +
                      Configuration::getKickableMargin(myself.getType());
    }
    int neededCycles = max(0, (int)(ceil(iP2i) - reachableArea)); // predictedCyclesToPoint
    if (neededCycles > 1) neededCycles++;
    for (int j = 1; j < neededCycles; ++j)
      dist += (bSpeed *= Configuration::ballDecay);
    if (dist <= distanceFromPointToPoint(p2, iP))
      smallestDistance = 0; //return true;
    else
      smallestDistance = min(smallestDistance, dist - distanceFromPointToPoint(p2, iP)); 
  }
  return smallestDistance * 20; // just to get bigger easier to compare values
}


//*********************************************************************************
//new isIntersected function
//contains 2 tolerance constants:
//width of pass line
//and
//stray

// normal: p1 = from, p2 = to; used: p1 = to, p2 = from
bool WorldModel::isIntersected(const Vector2D &p1, const Vector2D &p2) {
  
  static vector<Player>::iterator i;
  // float distP1P2 = distanceFromPointToPoint(p1, p2);
  float distP1P_i;
  float distPiP2;  
  float distP2P_i;
  Line line = Line::makeLineFromTwoPoints(p1,p2);
  
  //-8<- for each opponent check if he is too close to the pass line 
  for (i = theirTeam.getPlayers().begin(); i < theirTeam.getPlayers().end(); ++i) {
    if (!i->isGoodConfident(&time)){
	continue;
    }
    //distance from p1 to pi s projection on the line p1-p2
    static Vector2D pI;
    pI = line.getPointOnLineClosestTo(i->getPosition());

    distP1P_i = distanceFromPointToPoint(p1, pI);
    distP2P_i = distanceFromPointToPoint(p2, pI);
    //distance from pi to p2
    distPiP2 = distanceFromPointToPoint(i->getPosition(), p2);

    if (line.isInBetween(pI, p1, p2)) {
      if (line.distanceToPoint(i->getPosition()) <= enemyIntersectionTolerance(distP1P_i)) {
	return true;
      }
    }

  }
  //->8-
#ifdef CHECKTEAMMATEINTERSERCTION   
  //-8<- for each teammmate check if he is too close to the pass line
  for (i = myTeam.getPlayers().begin(); i < myTeam.getPlayers().end(); ++i) {
    /*
     * if i is neither at p1 nor at p2
     * and i is in between p1 and p2 and is too close to its projection on the pass line 
     */
    if(*i == myself){
      continue;
    }
    if (!i->isGoodConfident(&time)) {
      continue;
    }
    if ( (i->getPosition() != p1) && (i->getPosition() != p2) ) {
      if (line.isInBetween(i->getPosition(), p1, p2)) {
	if (line.distanceToPoint(i->getPosition()) <= teammateIntersectionTolerance(distP1P_i)) {
	  return true;
	}
      }
    }
  }
  //->8-
#endif
  return false;
}


/*************************************************************************/

float WorldModel::enemyIntersectionTolerance(float distance) {
   //decrease of one or both of these may result in higher probability to lose the ball
  const float EPASSWIDTH = 3;
  const float ESTRAY = 0.3;
  
  return ( distance * ESTRAY + EPASSWIDTH );
}

/*************************************************************************/

float WorldModel::teammateIntersectionTolerance(float distance) {
  //for teammates we may assume that they wont intentionally mess up our pass,
  //so we can tolerate closer values for them
  const float TPASSWIDTH = 0.5;
  const float TSTRAY = 0.1;
  
  return ( distance * TSTRAY + TPASSWIDTH );
}

/*************************************************************************/

bool WorldModel::isNotMarked(const Player &teammate){
  //Player *opp = closestOpponentToPlayer(teammate);
  const Player *opp = closestInSetToPlayer(PS_Opponents, teammate);
  if (opp != NULL){
    if (distanceFromPlayerToPoint(teammate, opp->getPosition()) >
        IS_ATTACKED_RADIUS)
      return true;
    else
      return false;
  }
  else
    return true;
}

/*************************************************************************/

bool WorldModel::isWellPositioned(const Player & passcand){
  if(passcand == *wellPositionedPlayer())
    return true;
  else
    return false;
}

/*************************************************************************/

bool WorldModel::isUnvulnerable(const Player & passcand){
  if( passcand == *mostUnvulnerablePlayer())
    return true;
  else
    return false;
}

/*************************************************************************/

const Player* WorldModel::wellPositionedPlayer(){
  // Ab durch die Mitte:
  Vector2D oppGoal(52, 0);
  //return closestTeammateToPoint(oppGoal);
  return closestInSetToPoint(PS_Teammates, oppGoal);
}

/*************************************************************************/

const Player* WorldModel::mostUnvulnerablePlayer(){
  /* 
   * Zu bedenken ist, dass der Spieler ausgesucht wird mit der groessten Distanz
   * zum Opp, bzw. der ohne Opp. Dummerweise gilt h.w.d. die eingeschraenkte Wahrnehmung.
   */
  float Dist = -2;
  float tmpDist = 0;
  static vector<Player>::iterator i, j;
  //vector<Player> &opponents = /*situation.*/theirTeam.getPlayers();
  //vector<Player> &teammates = /*situation.*/myTeam.getPlayers();

  const Player *unvulnerableOne = NULL;

  for( i = teammates->begin(); i != teammates->end(); ++i ){
    for(j = opponents->begin(); j != opponents->end(); ++j){
      if (opponents->size() == 0){
        unvulnerableOne = (const Player *)&(*i);
        return unvulnerableOne;
      }
      tmpDist = distanceFromPlayerToPoint(*i, j->getPosition());
    }
    if(tmpDist > Dist){
      Dist = tmpDist;
      unvulnerableOne = (const Player *)&(*i);
    }
  }
  return unvulnerableOne;
}

/*************************************************************************/


bool WorldModel::getPassPoint(Vector2D &passPoint,
			      Vector2D &wishedVelocity) {
  //calculates a point an passes there
  static vector<Player>::iterator i;
  /*
  int diff=0;
  timeval begin;
  timeval end;
  gettimeofday(&begin,NULL);
  */

  static Vector2D testDir;
  static Vector2D ourPos;
  static Vector2D fromMeToGoal;
  
  Vector2D ballVel = ball.getVelocity();
  float theirDistance, difference, ourDistance;
  ourDistance = theirDistance = UNDEFINED_FLOAT;
  static Vector2D target, tempTarget;
  static Vector2D wishedVelocitySelected;
  bool offensive = true;
  bool firstFound = false;
  
  Player *ourPassPartner = NULL;
 

  //verhalten 
  //do not pass in defense, because its too dangerous
  if(ball.getPositionPt()->getX() < -10)
    return false;

  fromMeToGoal = Vector2D(52,0) - myself.getPosition();
  int angle = (int)fromMeToGoal.getDirection();  
    //discretesize field in front of us and calculate for each point, if 
  //a teamate is clearly the closest to this position
  
  for(int distance = 9; distance <= 30; distance += 3){
    //for(int degree = -90; degree <= 90; degree += 4) {
    for(int degree = angle-90; degree <= angle+90; degree += 6) {
      // pass vector
      ;
      ourDistance = theirDistance = UNDEFINED_FLOAT; 
      testDir.setVector2D(distance, degree, POLAR);
      
      // APL_POINT_MOSH(ball.getPosition() + testDir);
    
      //dont pass if agent has not been looking in direction of target for
      //the last three see`s
      if(fabs(degree - directionHistory[0]) > 35 &&
  	 fabs(degree - directionHistory[1]) > 35 &&
	 fabs(degree - directionHistory[2]) > 35)
	 continue;
      tempTarget= ball.getPosition() + testDir;      

      //avoid passing out of the field (with tolerance because of noise)
      if((tempTarget).getY() < (-30 + (distance / 10)) ||
	 (tempTarget).getY() > ( 30 - (distance / 10)) ||
	 (tempTarget).getX() > ( 50 - (distance / 10))  )
	continue;
      
      //Test our Players (how far we are away from passpoint)
      for (i =  myTeam.getPlayers().begin(); 
	   i != myTeam.getPlayers().end(); 
	   ++i) {
	//	if(i->getPosition()!=NULL)
	if(i->isHighConfident(getTime()))
	  if(i->getPositionPt()->getX() > ConfigFile::backPassBorder || 
	     i->getPositionPt()->getX() > myself.getPositionPt()->getX()){
	    if (i->getUniformNumber() == myself.getUniformNumber()) continue;
	    if ( i->getPosition().getX() 
		 > getOffsideX(ConfigFile::goodConfidencePlayer) ) continue;
	    
	    float pDist = (i->getPosition() 
			   + i->getVelocity() ).distanceTo(tempTarget);
	    if(i->isGoalKeeper())
	      pDist = ((i->getPosition() 
			+ i->getVelocity()).distanceTo(tempTarget)
		       - Configuration::catchableAreaL);
	    
	    
	    if(pDist < ourDistance){
	      ourDistance = pDist;
	      ourPos = i->getPosition();
	      ourPassPartner = &(*i);
	    }    	    
	  }
      }

      if (ourPassPartner == NULL) continue;
      passPointUnumToSay = ourPassPartner->getUniformNumber();
      
      //Test their Players
      for (i = theirTeam.getPlayers().begin(); 
	   i < theirTeam.getPlayers().end(); 
	   ++i) {
	//if(i->getPosition()!=NULL)
	if(ball.getPositionPt()->getX() <= i->getPositionPt()->getX() + 30.0){
	  
	  float pDist = ((*(i->getPositionPt())) + i->getVelocity()).distanceTo(tempTarget);
	  
	  if(pDist < theirDistance)
	    theirDistance = pDist;
	} 
      }
      
      //set Goalie dummy,because he may not be seen
      // !!! should first check if he is seen with isVisible()
      /*
      if(Vector2D(43,0).distanceTo(tempTarget)<=theirDistance)
	theirDistance=Vector2D(43,0).distanceTo(tempTarget);
      */

      //if one of our players is in a suitable position for this pass point
      if(ourDistance + 5 <= myself.getPosition().distanceTo(tempTarget))
	if(ourDistance < (theirDistance - ( 2 + (testDir.getLength()/10))/*(3*testDir.getLength()/5)*/) 
	   && Field::FIELD.isInside(tempTarget)
	   ){
	 
	  /* calculates how fast i can pass while still be able to
	     intercept ball at passPoint (ball and player should be at 
	     passpoint at nearly the same time)*/

	  // cycles needed to go to passpoint (+ buffer for lame agents)
	  // int cycles=(int)(ourDistance/1.0/*Configuration::playerSpeedMax*/)+2;
	 
	  int turnCycles = int(testDir.getLength()/5);
	  if (ourPassPartner == NULL) continue;
	  int cycles = (ourPassPartner->predictedCyclesToPoint(tempTarget)+3
			+ turnCycles);   
	 
	  float bdW = 0;
	  float factorW = 1 + Configuration::ballDecay;
	  //calculation factor=1+(bd)+()+()...()
	  for(int h = 2; h <= cycles; ++h){
	    bdW = Configuration::ballDecay;    
	    for(int l = 1; l < h; ++l){
	      bdW *= Configuration::ballDecay;
	    }
	    factorW += bdW;
	  }
	  //velocity at kicking the ball
	  float velLenghtW = testDir.getLength() / factorW;

	  static Vector2D relBallPos;
	  float ballDist,ballAngle;
	  //calculateKickRate related
	  relBallPos = ball.getPosition() - myself.getPosition();
	  ballDist   = relBallPos.getLength();
	  ballAngle  = angleToBallRel2Body();

	  static Vector2D wishedAccel;
	  static Vector2D wishedVel;
	  wishedVel = testDir;
	  wishedVel.setLength(velLenghtW);
	  if (!ballVel.isInvalid())
	    wishedAccel = wishedVel - ballVel;
	  else
	    wishedAccel = wishedVel;
	 	 
	  //calculate the needed power to kick that fast
	  float neededPower = wishedAccel.getLength() 
	    / ball.calculateKickRate(ballDist, ballAngle)
	    / Configuration::kickPowerRate;
	  
	  //if power not reachable try shorter velocity
	  while(neededPower > 100 && wishedVel.getLengthSq()>0.0025){
	    wishedVel.setLength(wishedVel.getLength()-0.05);
	    if (!ballVel.isInvalid())
	      wishedAccel = wishedVel - ballVel;
	    else
	      wishedAccel = wishedVel;
	    neededPower = wishedAccel.getLength() 
	      / ball.calculateKickRate(ballDist, ballAngle)
	      / Configuration::kickPowerRate;
	  }

	  /*
	  //velocity at pass point
	  float velRestLengthW = wishedVel.getLength();	  
	  for (int h=1;h<cycles;h++)
	    velRestLengthW *= Configuration::ballDecay;
	 
	  float speed = calcKickSpeedToTravel( distance, velRestLengthW);
	  Vector2D predBallVelocity;
	  predBallVelocity.setVector2D(speed,degree,POLAR);
	  */
	  // with the now calculated speed we can test if no enemy is able
	  // to intercept the ball until he reaches passpoint 
	  
	  if(isNotIntersected_PIOS(wishedVel, ball.getPosition(),
				   tempTarget, ourPos)){
	 	    
	    if(!firstFound){
	      //first target found    
	      target = tempTarget;
	       difference = theirDistance - ourDistance;
	       wishedVelocitySelected = wishedVel;
	       firstFound = true;
	    }else{ //change target if our agent is more free
	      if(!offensive){//safer method to pass
		for (i = myTeam.getPlayers().begin(); 
		     i < myTeam.getPlayers().end(); 
		     ++i)
		  if(i->isHighConfident(&time))
		    if(target.distanceToSq(i->getPosition()) >
		       (tempTarget).distanceToSq(i->getPosition())){  
		      difference = theirDistance - ourDistance;
		      target = tempTarget;
		      wishedVelocitySelected = wishedVel;
		    }
	      }else//else more offensive and more risky passPoint
		if(target.distanceToSq( Vector2D(52,0) ) >
		   (tempTarget).distanceToSq( Vector2D(52,0) )){  
		  difference = theirDistance - ourDistance;
		  target = tempTarget;
		  wishedVelocitySelected = wishedVel;
		}
	    }
	  }
	}      
    } //-----\ loop over possible
  }   //-----/     passpoints
  
  
  //gettimeofday(&end,NULL);
  //diff = timerDiffMsec(&end, &begin); // millis  
  //  DEBUG2_MOSH("----> passing lasts :",diff);
  
  //No acceptable passpoint where found
  if(!firstFound){
    return false;
  }else{
    wishedVelocity = wishedVelocitySelected;
    passPoint = target;
    
    return true;
  }
}
/*************************************************************************/


bool WorldModel::canDribbleBehindDefense(Vector2D &passPoint, Vector2D &wishedVelocity) {
  return false;
  static vector<Player>::iterator i;
 
  //APL_ELLIPSE3_MOSH(Vector2D(2,0),0.5,0.5, (APLLevel) 4  , 255, 255, 255);    
  static Vector2D testDir;
  static Vector2D myPos;
  static Vector2D fromMeToGoal;
  
  Vector2D ballVel = ball.getVelocity();
  float theirDistance, myDistance,shortestDist;
  myDistance = theirDistance = UNDEFINED_FLOAT;
  static Vector2D target, tempTarget;
 
  //bool offensive = true;
  bool firstFound = false;
  shortestDist=0; 
  Player *ourPassPartner = NULL;
 
  APL_LINE_MOSH(Vector2D(getOffsideX(ConfigFile::goodConfidencePlayer)-0.3,32)          , Vector2D(getOffsideX(ConfigFile::goodConfidencePlayer)-0.3,-32));  
  
  //near goal use normal dribble 
  if(ball.getPositionPt()->distanceTo(Vector2D(52,0)) < 32)
    return false;
  
  fromMeToGoal = Vector2D(52,0) - myself.getPosition();
  int angle = (int)fromMeToGoal.getDirection();  
    //discretesize field in front of us and calculate for each point, if 
  //a teamate is clearly the closest to this position
  
  /* 
  if(ball.getPosition().getX() 
     <= getOffsideX(ConfigFile::goodConfidencePlayer)-0.3)
    return false;
  //Ich Bin Durch !!! JUHU
  */
  

  for(double distance = 5 ; distance <= 15; distance += 2 ){
    for(int degree = angle-20; degree <= angle + 20; degree += 10) {
      // vorleg vector
      myDistance = theirDistance = UNDEFINED_FLOAT; 
      testDir.setVector2D(distance, degree, POLAR);
      
          
      
      //dont vorleg if agent has not been looking in direction of target for
      //the last three see`s
      if(fabs(degree - directionHistory[0]) > 20 &&
  	 fabs(degree - directionHistory[1]) > 20 &&
	 fabs(degree - directionHistory[2]) > 20)
	 continue;

      
      tempTarget= ball.getPosition() + testDir;      

      if(tempTarget.getX() 
	 < getOffsideX(0.88) )
	continue;  
      
      //avoid vorleg to  
      if((tempTarget).getY() < (-30 + (distance / 5)) ||
	 (tempTarget).getY() > ( 30 - (distance / 5)) ||
	 (tempTarget).getX() > ( 48 - (distance / 5))  )
	continue;
      
      //Test myself (how far we are away from passpoint)
      
      float pDist = (myself.getPosition() + myself.getVelocity() ).distanceTo(tempTarget);	    
      
      myDistance = pDist;
      myPos = myself.getPosition();
      
      //Test their Players
      for (i = theirTeam.getPlayers().begin(); 
	   i < theirTeam.getPlayers().end(); 
	   ++i) {
	if(ball.getPositionPt()->getX() <= i->getPositionPt()->getX() + 30.0){
	
	  if (i->getConfidence(&time) < 0.88)
	    continue;
	  float pDist = ((*(i->getPositionPt())) + i->getVelocity()).distanceTo(tempTarget);
	  
	  //keep ball away from goalie to keep chance to score goal
	  if(i->isGoalKeeper())
	    pDist = ((*(i->getPositionPt())) + i->getVelocity()).distanceTo(tempTarget)-(3*Configuration::catchableAreaL);
	  
	  
	  if(pDist < theirDistance)
	    theirDistance = pDist;
	} 
      }
      //set Goalie dummy,because he may not be seen
      // !!! must be there, else we kick alway keeper not seen 
      //to hard to goal and keeper takes ball 
      
      if( (Vector2D(48,0).distanceTo(tempTarget) -(4*Configuration::catchableAreaL) ) < theirDistance )
	theirDistance=(Vector2D(43,0).distanceTo(tempTarget)
		       -(4*Configuration::catchableAreaL));
      
      
      
      //only if we are closer than enemy
      if(myDistance < (theirDistance - (2 +( testDir.getLength()/5 )) )
	 && Field::FIELD.isInside(tempTarget)
	 ){
	
	int cycles = int(testDir.getLength()/7)+1+(myself.predictedCyclesToPoint(tempTarget));   
	float bdW = 0;
	float factorW = 1 + Configuration::ballDecay;
	//calculation factor=1+(bd)+()+()...()
	for(int h = 2; h <= cycles; ++h){
	  bdW = Configuration::ballDecay;    
	  for(int l = 1; l < h; ++l){
	    bdW *= Configuration::ballDecay;
	  }
	  factorW += bdW;
	}
	//velocity at kicking the ball
	float velLenghtW = testDir.getLength() / factorW;

	static Vector2D relBallPos;
	float ballDist,ballAngle;
	
	//calculateKickRate related
	relBallPos = ball.getPosition() - myself.getPosition();
	ballDist   = relBallPos.getLength();
	ballAngle  = angleToBallRel2Body();
	
	static Vector2D wishedAccel;
	static Vector2D wishedVel;
	wishedVel = testDir;
	wishedVel.setLength(velLenghtW);
	if (!ballVel.isInvalid())
	  wishedAccel = wishedVel - ballVel;
	else
	  wishedAccel = wishedVel;

	float neededPower = wishedAccel.getLength(); 
	if(neededPower >100)
	  continue;
	 
	float distanceToGoal =tempTarget.distanceTo(Vector2D(52,0));
        
	if(!firstFound){//first target found    
	  target = tempTarget;
	  shortestDist=distanceToGoal;
	  firstFound = true;
	  wishedVelocity=wishedVel;
	  
	}else //change target if scoring Probability is higher 
	  if(distanceToGoal < shortestDist){//than the actual one
	    shortestDist=distanceToGoal;	      
	    target = tempTarget;
	    wishedVelocity=wishedVel;
	  }
      } //else try next point==>
      
    } //-----\ loop over possible
  }   //-----/     vorlegpoints
  
  //No acceptable VORLEG_Point where found
  if(!firstFound){
    return false;
  }else{
    DEBUG1_MOSH("vorleg");
    passPoint = target;
    APL_ELLIPSE3_MOSH(passPoint,0.5,0.5, (APLLevel) 4  , 255, 255, 255);    
    return true;
  }
}

/*************************************************************************/

const Player* WorldModel::getPassCand(bool offensive, bool considerOffside, bool considerBackPass, PassType pt){
  // unused parameter:
  offensive = offensive;
  
  Player *passCand = NULL;
  static vector<Player>::iterator i;

  // has to be a positive value
  // only pass to players who are 5m (own half) or 2m (opp half) in 
  // front of me
  float xThreshold = ball.getPositionPt()->getX() < 0 ? 5 : 2;
  float minDist = BIG_FLOAT;
  float maxX = 0;
  float maxOppDist = 0;
  float ownX = myself.getPositionPt()->getX();

  for (i = myTeam.getPlayers().begin(); i < myTeam.getPlayers().end(); ++i) {
    // do not pass to myself
    if (i->getUniformNumber() == myself.getUniformNumber())
      continue;
    // do not pass to goalie
    if (i->getUniformNumber() == 1)
      continue;
    // do not pass to players you haven't seen for some time
    if(!i->isHighConfident(&time)) 
      continue;
    // dont pass to players who are offside
    if (considerOffside && isOffside(i->getPosition()))
      continue;
    // ignore players who are outside the field
    if (!field.FIELD.isInside(i->getPosition()))
      continue;
    // dont pass do players who are far behind me
    if (considerBackPass && ball.getPositionPt()->getX() < ConfigFile::backPassBorder){
      float otherX = i->getPositionPt()->getX();
      if((ownX + xThreshold) > otherX)
	continue;
    }
    // dont pass to players who are marked
    float kS = calcKickSpeedToTravel(i->getPositionPt()->distanceTo(ball.getPosition()), ConfigFile::passEndSpeed );
    if(isIntersected_NH(i->getPosition(), ball.getPosition(), kS ) )
      continue;

    // The pass to this candidate would not be intercepted and
    // he is in a good enough position. Now find the one with
    // the best position.
    float hisX         = i->getPositionPt()->getX();
    float dist         = myself.getPositionPt()->distanceToSq(i->getPositionPt());
    const Player *closestOpp = closestInSetToPoint(PS_Opponents, i->getPositionPt());

    if (pt == PT_DEFENSIVE){
      if (dist < minDist) {
	minDist = dist;
	passCand = &(*i);
      }
    }else if(pt == PT_OFFENSIVE){
      if (hisX > maxX){
	maxX = hisX;
	passCand = &(*i);
      }
    }else if(pt == PT_SECURE){
      if(closestOpp != NULL){
        float oppDist = distanceToPoint(closestOpp->getPosition());
	if (oppDist > maxOppDist){
	  maxOppDist = oppDist;
	  passCand = &(*i);
	}
      }
    }
  }
  return passCand;
}

/*************************************************************************/

float WorldModel::getFreeAngle(const Vector2D &start, 
			       const Vector2D &leftBorderPoint, 
			       const Vector2D &rightBorderPoint){
  static vector<float> angles;
  angles.clear(); // Just to be sure
  static vector<float>::const_iterator j;
  static vector<float>::const_iterator k;
  static vector<Player>::const_iterator i;
  static const float limit = 40 * 40; // ignore players that are further away than this
  float freeAngle = 0;
  float maxDist  = - BIG_FLOAT;
  float tempAngle;
  //transform borderPoints into borderAngles 
  float  leftBorder  = angleFromPointToPoint(start, leftBorderPoint);//field.getMarkerPosition(FLAG_RT30)
  float  rightBorder = angleFromPointToPoint(start, rightBorderPoint);//field.getMarkerPosition(FLAG_RB30)
  
  angles.push_back(leftBorder);
  angles.push_back(rightBorder);
  for(i = theirTeam.getPlayers().begin(); i != theirTeam.getPlayers().end(); ++i){
    if (!i->isGoodConfident(&time)) continue;
    /*
    Vector2D playerPos = i->getPosition();
    if (playerPos.getX() < leftBorderPoint.getX() &&
	playerPos.getX() < rightBorderPoint.getX()) continue;
    */
    tempAngle = angleFromPointToPoint(start, i->getPosition());
    bool inInterval = isAngInInterval(tempAngle, leftBorder, rightBorder);

    // DEBUG
    //bool inOld = (tempAngle > leftBorder) && (tempAngle < rightBorder);
    /*
    if (inInterval != inOld){
      DEBUG4_HARLEKIN("TEST: Unterschied, new: ", inInterval, " old: ", inOld);
    }else{
      DEBUG4_HARLEKIN("TEST: Kein Unterschied, new: ", inInterval, " old: ", inOld);
    }
    */

    if (inInterval){
      if (start.distanceToSq(i->getPosition()) < limit){
	angles.push_back(tempAngle);
      }
    }
    /** old method
    if((tempAngle > leftBorder) && (tempAngle < rightBorder) && 
       (start.distanceToSq(i->getPosition()) < limit)){
      angles.push_back(tempAngle);
    }
    */
  }
  sort(angles.begin(), angles.end());
  k = angles.begin();
  for(j = angles.begin(); ++k < angles.end(); ++j){
    float dist = getAbsAngleDifference(*j, *k);
    // DEBUG
    //float dist2 = fabs(*j - *k);
    //DEBUG4_HARLEKIN("TEST: new: ", dist, " old: ", dist2);
    if (dist > maxDist){
      maxDist = dist;
      freeAngle = (*j + maxDist/2.0f);
    }
  }
  return(freeAngle);
}
/*************************************************************************/
/* Checks if their team will get the ball -> so they got possession */
bool WorldModel::hasBallPossessionOpp(){
  const Player *fastest_TheirTeam = fastestInSetToBall(PS_Opponents);
  const Player *fastest_All = fastestInSetToBall(PS_Players);
  if(fastest_TheirTeam==fastest_All)
    return true;
  else
    return false;
}
/*************************************************************************/

/*************************************************************************/
/* Checks if our team will get the ball -> so we got possession */
bool WorldModel::hasBallPossession(){
  return hasBallTeam();
  // TODO: WAY TO EXPENSIVE!!!
  
  //const Player *fastest = fastestInSetToBall(PS_Players);
  // return (fastest != NULL && !fastest->isOpponent());
  
}
/*************************************************************************/

/* Checks if the specified player possesses the ball */
bool WorldModel::hasBallPlayer(const Player& player) const {
  static const float margin = BALL_POSSESSION_MARGIN * BALL_POSSESSION_MARGIN;
  return (isClosestToBall(player) &&
	  (player.getPosition().distanceToSq(ball.getPosition()) < margin));
}

/*************************************************************************/

/* Checks if the specified player has the ball in his kickableMargin */
bool WorldModel::isBallKickablePlayer(const Player& player) const {
  int type = player.getType();
  return ((distanceFromPlayerToBall(player) - Configuration::getPlayerSize(type)
           - Configuration::ballSize <= Configuration::getKickableMargin(type)));
}

/*************************************************************************/

/* Checks for every player in our team if he possesses the ball */
bool WorldModel::hasBallTeam() const {
  vector<Player>::const_iterator i;
  for (i = teammates->begin(); i != teammates->end(); ++i) {
    if (!i->isGoodConfident(&time)) continue;
    if (hasBallPlayer( *i )) return true;
  }
  return false;
}

/*************************************************************************/

/* Checks for every player in the opponent team if he possesses the ball */
bool WorldModel::hasBallOppTeam() const {
  static vector<Player>::const_iterator i;
  for (i = opponents->begin(); i != opponents->end(); ++i) {
    if (!i->isGoodConfident(&time)) continue;
    if (hasBallPlayer( *i )) return true;
  }
  return false;
}

/*************************************************************************/

/* Checks for every player in the agent's team if he has ball in his kickableMargin */
bool WorldModel::isBallKickableTeam() const {
  static vector<Player>::const_iterator i;
  for (i = teammates->begin(); i != teammates->end(); ++i) {
    if (!i->isGoodConfident(&time)) continue;
    if (isBallKickablePlayer( *i )) return true;    
  }
  return false;
}

/*************************************************************************/

/* Checks for every player in the opponent team if he has ball in his kickableMargin */
bool WorldModel::isBallKickableOppTeam() const {
  static vector<Player>::const_iterator i;
  for (i = opponents->begin(); i != opponents->end(); ++i) {
    if (!i->isGoodConfident(&time)) continue;
    if (isBallKickablePlayer( *i )) return true;
  }
  return false;
}

/*************************************************************************/

/* Checks for every player in the opponent team if he has ball in his kickableMargin */
const Player * WorldModel::getAnEnemyWithBallKickable() const {
  //bool isBallKickable = false;
  static vector<Player>::const_iterator i;
  for (i = opponents->begin(); i != opponents->end(); ++i) {
    if (!i->isGoodConfident(&time)) continue;
    if (isBallKickablePlayer( *i )){
      return &(*i);
    }
  }
  //return isBallKickable;
  return NULL;
}

/*************************************************************************/

const Player * WorldModel::getTMWithBall() {
  static vector<Player>::iterator i;
  for (i = myTeam.getPlayers().begin(); i < myTeam.getPlayers().end(); ++i){
    if (!i->isGoodConfident(&time)) continue;
    if(hasBallPlayer(*i)){
      return(&(*i));
    }
  }
  return(NULL);
}

/*************************************************************************/

const Vector2D WorldModel::findBestPassCandOfferPoint(const Vector2D &tmPos,
						     float freeAngle) const {
  Line lineOfFire = Line::makeLineFromPositionAndAngle( tmPos, freeAngle );

  return lineOfFire.getPointOnLineClosestTo( myself.getPosition() );
}

/*************************************************************************/
/*** Methods concering tackling ***/

const double WorldModel::getTackleFailProb(){
  Vector2D player2ball = (ball.getPosition() - myself.getPosition()).rotate(-myself.getBodyDir());
  float tackle_dist;
  if (player2ball.getX() > 0.0){
    tackle_dist = Configuration::tackleDist;
  }else{
    tackle_dist = Configuration::tackleBackDist;
  }
  double failProb = pow( ( player2ball.getX() / tackle_dist  ), 
			 Configuration::tackleExponent) +
                    pow( ( player2ball.getY() / Configuration::tackleWidth ), 
			 Configuration::tackleExponent);
  return min(1.0, failProb);   
}

const bool WorldModel::isBallTackable(){
  double prob = getTackleFailProb();
  return prob <= ConfigFile::tackleThreshold;
}

/******************************************************************************/
/** This function positions the Goalie according to a found Pass Partner.
    It searches for a Pass Partner and gets the closest Goalie Kick Corner in
    relation to the Pass Partner. The Goalie will be moved to this point.
    @return a Vector2D which is the Point where the Goalie shall position itself.
    If no Pass Partner is found he positions himself to the left side (very random).
    @todo this is just a temporary design, it is still much too easy. But still better
    than before, because no passes in front line of the goal will be made.
*/
Vector2D WorldModel::findGoalieKickOffPoint(){
 
  const Vector2D *leftPoint  = &(Field::GOALIE_KICK_OFF_POINT_LEFT);
  const Vector2D *rightPoint = &(Field::GOALIE_KICK_OFF_POINT_RIGHT);


  Formation *f = Formations::getCurrentFormation();

  const Player* cand = NULL;
  float maxDist = 8.0;

  static vector<Player>::iterator i;
  for (i = teammates->begin(); i != teammates->end(); ++i){
    if (!i->isGoodConfident(&time)) continue;
    int unum = i->getUniformNumber();
    if (unum == 1) continue;
    int bID = f->getBehaviour(unum);
    // Only pass to WingDefenders
    if (bID != B_WingDefenderLeft &&
	bID != B_WingDefenderRight){
      continue;
    }
    float dist = min((i->getPosition() - *leftPoint).getLength(),
		     (i->getPosition() - *rightPoint).getLength());
    if (isIntersected_NH(i->getPosition(), *leftPoint, 
			 calcKickSpeedToTravel(dist, ConfigFile::passEndSpeed)))
      continue;
    if (dist > maxDist){
      cand = &(*i);
      maxDist = dist;
    }
  }

  if (cand != NULL)
    return (cand->getPositionPt()->getY() > 0 ? *rightPoint : *leftPoint);

  return (drand48() > 0.5 ? *rightPoint : *leftPoint);

}


/*************************************************************************/
/* the following methods try to consider if there is a promising point
   to shoot the ball to in order to score a goal */
/*************************************************************************/


/*******************************************************************************/
/**
 * This method passes through a number of points, which is given from the 
 * ConfigFile and checks the probability for each point. It takes the most
 * promising one and checks if it has enough probability for scoring a goal.
 * @param  from, point from there to shoot at the goal
 * @return valid point with a good probability to score a goal on
 *         invalid point, otherwise...
 */
/*******************************************************************************/

Vector2D WorldModel::getBestPointToShootTo(const Vector2D &from, float minSpeed){
   // use static vectors so we dont need to 
   // create new ones every time
  static Vector2D best;
  static Vector2D current;
  double bestValue = -1.0;
  double bestValueToGoal = -1.0;
  double stepSize  = Configuration::goalWidth;
  //stepSize /= ConfigFile::scoringProbabilitySteps;
  double goalX     = Field::OPP_GOAL_OUTLINE;
  
  float  goalY     = (Configuration::goalWidth-1)/2.0;
  // float  goalY     = (Configuration::goalWidth)/2.0;
  stepSize  = goalY/6;
  
  for (double i = -goalY; i <= goalY; i += stepSize){
 
    current.setVector2D(goalX, i);
    if ((current-from).getLengthSq() > 1024 ) 
      continue; 
    
    // get probability for hitting the goal
    double probGoal   = probToVector(from, current);
    if(probGoal <= 0.5)//if chance to low dont shoot to point
      continue;
    
    // get probability for passing the goalie and other enemys
    double probGoalie = getProbForPassingGoalie(from, current, minSpeed);    
    if(probGoalie == -1.0)
      continue;

    // select Point with best chance to pass
    /*
    if (probGoalie > bestValue){
      DEBUG1_MOSH("selected");
      bestValue = probGoalie;
      best = current;
    }
    */

    if (probGoalie > bestValue || probGoalie > 2){
      if(probGoalie >2 && bestValue >2){
	if(bestValueToGoal < probGoal){
	  DEBUG1_MOSH("selectedTOGOAL");	  
	  bestValue = probGoalie;
	  bestValueToGoal = probGoal;
	  best = current;
	}
      }
      else{
	DEBUG1_MOSH("selectedINTERSECTED");
	bestValue = probGoalie;
	bestValueToGoal = probGoal;
	best = current;
      }
    }
  }

  //return only a point if it has a good probability.
  if (bestValue != -1.0){
    DEBUG1_MOSH("shoot");//ConfigFile::scoringProbability)
    return best;
  }
  // otherwise return an invalid vector
  return Vector2D(NAN, NAN);
};

Vector2D WorldModel::getClosestPointToShootTo(const Vector2D &from, 
					      float minSpeed){
  // use static vectors so we dont need to 
  // create new ones every time
  ;
  static Vector2D closest;
  static Vector2D current;
  double bestDist = BIG_FLOAT;
  double stepSize  = Configuration::goalWidth - 0.2f;
  //stepSize /= (ConfigFile::scoringProbabilitySteps - 1);
  double goalX     = Field::OPP_GOAL_OUTLINE;
  float  goalY     = (Configuration::goalWidth-0.2f )/2.0f;
  stepSize  = goalY/6.0f;
  //  DEBUG4_MOSH("CLOSEstepSize ",stepSize," und ",(-goalY+stepSize));
  for (double i = -goalY; i <= goalY+1; i += stepSize){
    current.setVector2D(goalX, i);
    float distSq = (current-from).getLengthSq();
    if (distSq > 1024) 
      continue; 
    // get probability for point in goal
    double value = getProbabilityForScoringGoal(from, current, minSpeed);
    // Only points with a got probability and are close are wanted
    if (value >= ConfigFile::scoringProbability * 100.0f){
	if (distSq < bestDist){
	    bestDist = distSq;
	    closest = current;
	}
    }
  }

  //return only a point if one was found. ;)
  if (bestDist != BIG_FLOAT)
    return closest;
  // otherwise return an invalid vector
  return Vector2D(NAN, NAN);
};


/*    *************    OLD METHOD  ******************
Vector2D WorldModel::getBestPointToShootTo(const Vector2D &from, float minSpeed){
  // use static vectors so we dont need to 
  // create new ones every time
  static Vector2D best;
  static Vector2D current;
  double bestValue = 0;
  double stepSize  = Configuration::goalWidth;
  //stepSize /= ConfigFile::scoringProbabilitySteps;
  double goalX     = Field::OPP_GOAL_OUTLINE;
  
  float  goalY     = (Configuration::goalWidth-0.5)/2.0;
  // float  goalY     = (Configuration::goalWidth)/2.0;
  stepSize  = goalY*2;
  DEBUG1_MOSH("stepSize");
  for (double i = -goalY; i <= goalY; i += stepSize){
  
    current.setVector2D(goalX, i);
    if ((current-from).getLengthSq() > 1024) // 32  
      continue; 
    // get probability for point in goal
    double value = getProbabilityForScoringGoal(from, current, minSpeed);
    if (value > bestValue){
      bestValue = value;
      best = current;
    }
  }
  //return only a point if it has a good probability.
  if (bestValue >= 0.45)//ConfigFile::scoringProbability)
    return best;
  // otherwise return an invalid vector
  return Vector2D(NAN, NAN);
};
*/


/*******************************************************************************/
/**
 * This method brings together the two probabilities, first the probability for
 * passing the goalie and secondly the probability, that the ball will end up 
 * in the goal anyway.
 * @param from, point from there to shoot at the goal
 * @param to, point to shoot to in the goal
 * @return probability for scoring a goal
 */
/*******************************************************************************/
double WorldModel::getProbabilityForScoringGoal(const Vector2D &from, 
						const Vector2D &to,
						float minSpeed){
  double probGoalie = getProbForPassingGoalie(from, to, minSpeed);
  double probGoal   = probToVector(from, to);
  DEBUG8_MOSH("getProbabilityForScoringGoal(", from, ",", to,"): ", probGoalie, " * ", probGoal);
  DEBUG1_MOSH(*getTime());
  return probGoalie * probGoal;
};


/*******************************************************************************/
/**
 * This method checks whether the ball can be intersected by the goalie
 * on the way to the point to shoot to, with maximum available ballSpeed
 * @param from, point from there to shoot at the goal
 * @param to, point to shoot to in the goal
 * @return currently 1.0, if the ball cant be intersected
 *         0.0, otherwise...
 */
/*******************************************************************************/
//Test Method

double WorldModel::getProbForPassingGoalie(const Vector2D &from, 
					   const Vector2D &to,
					   float minSpeed){
  double intersected = -1;
  //for own position use
  if (from == ball.getPosition()){
    //for power higher than max, the shooting vector is shorter but
    //in the same direction (with power = 100) 
    float shootDir  = angleFromPointToPoint(from, to);
    float ballDist  = distanceToBall();
    float ballAngle = angleToBallRel2Body();
    float speed     = 100.0 * Configuration::kickPowerRate *
	              ball.calculateKickRate( ballDist, ballAngle );
    
    float tmp       = ball.getVelocity().rotated( -shootDir ).getY();
    float angle     = normalizeTo180Deg(shootDir - asinDeg( tmp / speed ));
    angle           = normalizeTo180Deg( angle - myself.getBodyDir() );
    
    //estimation of the maximum ballSpeed hopefully in the correct direction ;)
    float ballSpeed = estimateBallVelocity(100.0, angle).getLength();
    if (minSpeed != -1 && minSpeed > ballSpeed){
      return 0.0f;
    }
    intersected = isIntersected_NH_Risk(to, from, ballSpeed);
    
  }else{
    //to use to consider the situation of an other player 
    intersected = isIntersected_NH_Risk(to, from, Configuration::ballSpeedMax);
  }
  
  return intersected;
  //return -1.0;
 
}
/************************ old method ********************************
double WorldModel::getProbForPassingGoalie(const Vector2D &from, 
					   const Vector2D &to,
					   float minSpeed){
  bool intersected = false;
  //for own position use
  if (from == ball.getPosition()){
    //for power higher than max, the shooting vector is shorter but
    //in the same direction (with power = 100) 
    float shootDir  = angleFromPointToPoint(from, to);
    float ballDist  = distanceToBall();
    float ballAngle = angleToBallRel2Body();
    float speed     = 100.0 * Configuration::kickPowerRate *
	              ball.calculateKickRate( ballDist, ballAngle );
    
    float tmp       = ball.getVelocity().rotated( -shootDir ).getY();
    float angle     = normalizeTo180Deg(shootDir - asinDeg( tmp / speed ));
    angle           = normalizeTo180Deg( angle - myself.getBodyDir() );
    
    //estimation of the maximum ballSpeed hopefully in the correct direction ;)
    float ballSpeed = estimateBallVelocity(100.0, angle).getLength();
    if (minSpeed != -1 && minSpeed > ballSpeed){
      return 0.0f;
    }
    intersected = isIntersected_NH(to, from, ballSpeed);
    
  }else{
    //to use to consider the situation of an other player 
    intersected = isIntersected_NH(to, from, Configuration::ballSpeedMax);
  }
  if(intersected)
    return 0.0f;
  else
    return 1.0;
}
*/
//NOT IN USE
/* OLD JELLE IMPLEMENTATION
  double WorldModel::getProbForPassinGoalie(Vector2D pos){
    const Player *p = theirTeam.getGoalie();
    if (p == NULL) {
      return 1.0;
    }
    Vector2D wayOfBall = pos - ball.getPosition();
    Vector2D wayToGoalie = p->getPosition() - ball.getPosition();
    double distance = wayToGoalie.getLength();
    double angle = fabs(normalizeTo180Deg(wayOfBall.getDirection() - wayToGoalie.getDirection()));
    if (distance < 3 || distance > 15) return 0.0;
    return (1 / (1 + exp(-9.5*getLinearDiscriminant(angle, distance))));
  };

  our ACTUAL Method

double WorldModel::getProbForPassingGoalie(const Vector2D &from, 
					   const Vector2D &to,
					   float minSpeed){
  bool intersected = false;
  //for own position use
  if (from == ball.getPosition()){
    //for power higher than max, the shooting vector is shorter but
    //in the same direction (with power = 100) 
    float shootDir  = angleFromPointToPoint(from, to);
    float ballDist  = distanceToBall();
    float ballAngle = angleToBallRel2Body();
    float speed     = 100.0 * Configuration::kickPowerRate *
	              ball.calculateKickRate( ballDist, ballAngle );
    //DEBUG4_HARLEKIN("Probs: ", speed, " : ", speed2);
    float tmp       = ball.getVelocity().rotated( -shootDir ).getY();
    float angle     = normalizeTo180Deg(shootDir - asinDeg( tmp / speed ));
    angle           = normalizeTo180Deg( angle - myself.getBodyDir() );
    
    //estimation of the maximum ballSpeed hopefully in the correct direction ;)
    float ballSpeed = estimateBallVelocity(100.0, angle).getLength();
    if (minSpeed != -1 && minSpeed > ballSpeed){
      DEBUG2_HARLEKIN("ScoreGoal::ball would be too slow: ", ballSpeed);
      return 0.0f;
    }
    intersected = isIntersected_NH(to, from, ballSpeed);
    //DEBUG6_HARLEKIN(" intersected: from ", from, " to ", to ,":", intersected);
    //DEBUG4_HARLEKIN("GoalieInf: Conf: ", theirTeam.getGoalie()->getConfidence(&time), " Position: ", theirTeam.getGoalie()->getPosition());
  }else{
    //to use to consider the situation of an other player 
    intersected = isIntersected_NH(to, from, Configuration::ballSpeedMax);
  }
  if (intersected)
    return 0.0f;
  else
    return 1.0f;
}
*/


/*******************************************************************************/
/**
 * This method calculates the probability that the ball will on his way
 * end up in the goal anyway. It uses at the moment the taylorFormel.
 * @param from, point from there to shoot at the goal
 * @param to, point to shoot to in the goal
 * @return probability, that ball will on his way end up in the goal
 */
/*******************************************************************************/
double WorldModel::probToVector(const Vector2D &from, 
				const Vector2D &to){
  static Line way = Line(0,0,0);
  static Vector2D p1, p2;
  way.updateLineFromTwoPoints(from, to);
  p1 = way.getPointOnLineClosestTo(Field::OPP_GOAL_POST_TOP_POS);
  p2 = way.getPointOnLineClosestTo(Field::OPP_GOAL_POST_BOTTOM_POS);
  double dist1 = p1.distanceTo(&Field::OPP_GOAL_POST_TOP_POS);
  double dist2 = p2.distanceTo(&Field::OPP_GOAL_POST_BOTTOM_POS);

  //calculation for both posts with the simpson formel
  double prob1 = simpsonFormel(dist1, 0, ConfigFile::simpsonSteps, getDeviationTo(from, p1), &(WorldModel::gaussian));
  double prob2 = simpsonFormel(dist2, 0, ConfigFile::simpsonSteps, getDeviationTo(from, p2), &(WorldModel::gaussian));

/* NOT IN USE - BECAUSE OF WRONG VALUES
  double prob2 = 1;
  //prob2 = taylorFormel(-dist1, dist2, 6, 0, NULL);
  prob2 -= taylorFormel(from, -dist1, dist2, ConfigFile::taylorSteps, p1, p2, &(WorldModel::gaussian));
*/
  //probability, that ball will on his way end up in the goal
  return prob1 + prob2;
};


/*******************************************************************************/
/**
 * This formula calculates the standard deviation of the ball velocity direction
 * over a given distance. This calculation was publicised by Jelle Kok(UvA) and 
 * is based on extensive testing.
 * @param from, point from there to shoot at the goal
 * @param to, point to shoot to in the goal
 * @return the standard deviation
 */
/*******************************************************************************/
  double WorldModel::getDeviationTo(const Vector2D &from, const Vector2D &to){
    return -1.88 * log(1.0f - (from.distanceTo(to) / 45.0f) );
  };


/*******************************************************************************/
/**
 * This method simply approximate - h - i dont know take the simpson formel 
 * anyway! - STOP DOCUMENTATION HERE
 */
/*******************************************************************************/
double WorldModel::taylorFormel(const Vector2D &from, double t1, double t2, 
				int d, const Vector2D &p1, const Vector2D &p2, 
				double (*f)(double x, double dev)){
  t1 = f(t1,getDeviationTo(from, p1));
  t2 = f(t2,getDeviationTo(from, p2));
 
  if (t1 > t2){
    double temp = t1;
    t1 = t2;
    t2 = temp;
  } 
  double value = 0;
  double nfak = 1;
  for (int n = 0; n < d; ++n){
    // calculate faculty
    if (n) nfak *= n; 
    double mult = 2*n+1;
    double tlp = pow(t2, mult);// * f(t2, dev);
    double trp = pow(t1, mult);// * f(t1, dev);
    double tp = tlp - trp;
    double p = tp / (mult * nfak);
    if (n%2 != 0) p *= -1;
    value += p;
  }
  return value;
}


/*******************************************************************************/
/**
 * This method calculates the integral value for the given values. We use it at
 * the moment only to consider which point in the goal is the most promising one!
 * @param l left integral border
 * @param r right integral border
 * @param d numbers of steps to use --> see ConfigFile
 * @param dev deviation of gaussian curve  
 * @param function for goal shot --> gaussian
 * @return value of integral
 */
/*******************************************************************************/
double WorldModel::simpsonFormel(double l, double r, int d, double dev, double (*f)(double x, double dev)){
  if (l == r) return(0.0);
  else if (l > r) {
    //linke > rechte Grenze --> vertauschen
    double swap = l;
    l = r;
    r = swap;
  }
  if (d < 4) d = 4;
  if (!(d % 2)) ++d; //gerade anzahl an sttzstellen schaffen
  double delta = ((r - l) / (double)d); //deltaX = (rechte Grenze - linke Grenze) geteilt durch Schrittzahl
  double A = 0.0;
  double x = l + delta;   //x = linke grenze + schritt (x1)
  //gerade Summen : Summe 3 = 4*(y2 + y4 + ... + y(2n-2))
  for(int i = 2; i < ((2*d)-2); i += 2) {
    A += f(x,dev);
    x += delta; //x 1 schritt weiter setzen
  }
  A *= 2.0;
  x = l + (0.5 * delta);   //x = linke grenze + schritt (x1)
  //ungerade Summen : Summe 2 = 2*(y1 + y3 + ... + y(2n-1))
  for(int i = 1; i < ((2*d)-1); i += 2) {
    A += f(x,dev);
    x += delta; //x 1 schritt weiter setzen
  }
  A *= 2.0;
  A += f(l,dev) + f(r,dev); //Summe 1 = y0 + y(2n)
  A *= (0.5 * delta / 3); //Fl?che = (Summe 1 + Summe 2 + Summe 3) * h / 3 (h = delta/2)
  return(A);
}


// $Source: /home/robocup/virtualwerder/CVSRepository/VirtualWerder2004/worldmodel/WorldModel_HighLevel.cc,v $
